import contextlib
import functools
import os
import random
import sys
import threading
import time
import unittest
import warnings
from concurrent.futures._base import (
    CANCELLED,
    CANCELLED_AND_NOTIFIED,
    FINISHED,
    PENDING,
    RUNNING,
)
from pathlib import Path

sys.path.append(
    os.fspath(Path(__file__).resolve().parent.parent.parent / "test")
)
import mpitestutil as testutil

from mpi4py import MPI, futures

SHARED_POOL = futures._core.SharedPool is not None
WORLD_SIZE = MPI.COMM_WORLD.Get_size()


def create_future(state=PENDING, exception=None, result=None):
    f = futures.Future()
    f._state = state
    f._exception = exception
    f._result = result
    return f


PENDING_FUTURE = create_future(state=PENDING)
RUNNING_FUTURE = create_future(state=RUNNING)
CANCELLED_FUTURE = create_future(state=CANCELLED)
CANCELLED_AND_NOTIFIED_FUTURE = create_future(state=CANCELLED_AND_NOTIFIED)
EXCEPTION_FUTURE = create_future(state=FINISHED, exception=OSError())
SUCCESSFUL_FUTURE = create_future(state=FINISHED, result=42)


def mul(x, y):  # noqa: FURB118
    return x * y


def sleep_and_raise(t):
    time.sleep(t)
    raise RuntimeError("this is an exception")


def check_global_var(x):
    return global_var == x  # noqa: F821


def check_run_name(name):
    return __name__ == name


def check_comm_workers():
    comm = futures.get_comm_workers()
    return comm.Get_size()


def sys_flags_get(name):
    return getattr(sys.flags, name)


class ExecutorMixin:
    worker_count = 2

    def setUp(self):
        self.t1 = time.time()
        try:
            self.executor = self.executor_type(max_workers=self.worker_count)
        except NotImplementedError as exc:
            self.skipTest(str(exc))
        self._prime_executor()

    def tearDown(self):
        self.executor.shutdown(wait=True)
        dt = time.time() - self.t1
        self.assertLess(dt, 60, "synchronization issue: test lasted too long")

    def _prime_executor(self):
        # Make sure that the executor is ready to do work before running the
        # tests. This should reduce the probability of timeouts in the tests.
        futures = [
            self.executor.submit(time.sleep, 0)
            for _ in range(self.worker_count)
        ]
        for f in futures:
            f.result()


class ProcessPoolMixin(ExecutorMixin):
    executor_type = futures.MPIPoolExecutor

    if "coverage" in sys.modules:
        executor_type = staticmethod(
            functools.partial(
                executor_type,
                python_args="-m coverage run".split(),
            )
        )


@unittest.skipIf(not SHARED_POOL, "not-shared-pool")
class SharedPoolInitTest(unittest.TestCase):
    executor_type = futures.MPIPoolExecutor

    def test_initializer_0(self):
        executor = self.executor_type(
            initializer=time.sleep,
            initargs=(0,),
        )
        executor.bootup()
        executor.submit(time.sleep, 0).result()
        executor.shutdown()

    def test_initializer_1(self):
        for _ in range(2):
            executor = self.executor_type(
                initializer=sleep_and_raise,
                initargs=(0.2,),
            )
            executor.submit(time.sleep, 0).cancel()
            future = executor.submit(time.sleep, 0)
            with self.assertRaises(futures.BrokenExecutor):
                executor.submit(time.sleep, 0).result()
            with self.assertRaises(futures.BrokenExecutor):
                future.result()
            with self.assertRaises(futures.BrokenExecutor):
                executor.submit(time.sleep, 0)

    def test_initializer_2(self):
        executor = self.executor_type(
            initializer=time.sleep,
            initargs=(0,),
        )
        executor.bootup()
        executor.submit(time.sleep, 0).result()
        executor.shutdown()

    def test_initializer_3(self):
        executor = self.executor_type()
        executor.submit(time.sleep, 0).result()
        executor.shutdown()

    def test_initializer_4(self):
        rnd = random.random

        def test(tid):
            with self.executor_type(
                initializer=time.sleep,
                initargs=(rnd() / 100,),
            ) as executor:
                futures.as_completed([
                    executor.submit(time.sleep, rnd() / 100)
                    for _ in range(executor.num_workers + tid)
                ])

        ts = [threading.Thread(target=test, args=(i,)) for i in range(5)]
        for t in ts:
            t.start()
        for t in ts:
            t.join()


class ProcessPoolInitTest(ProcessPoolMixin, unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _prime_executor(self):
        pass

    def test_init(self):
        self.executor_type()

    def test_init_args(self):
        self.executor_type(1)

    def test_init_kwargs(self):
        executor = self.executor_type(
            python_exe=sys.executable,
            max_workers=None,
            mpi_info={"soft": "0:1"},
            globals=None,
            main=False,
            path=[],
            wdir=Path.cwd(),
            env={},
            use_pkl5=None,
            backoff=0.001,
        )
        futures = [
            executor.submit(time.sleep, 0) for _ in range(self.worker_count)
        ]
        for f in futures:
            f.result()
        executor.shutdown()

    def test_init_pyargs(self):
        executor_type = futures.MPIPoolExecutor
        executor = executor_type(python_args=["-B", "-Wi"])
        executor.submit(time.sleep, 0).result()
        executor.shutdown()

    @unittest.skipIf(SHARED_POOL, "shared-pool")
    def test_init_sys_flags(self):
        executor_type = futures.MPIPoolExecutor
        sys_flags = [
            ("debug", "-d", True),
            ("optimize", "-O", 1),
            ("optimize", "-OO", 2),
            ("dont_write_bytecode", "-B", True),
            ("dev_mode", "-Xdev", True),
            ("utf8_mode", "-Xutf8", True),
        ]
        if sys.version_info >= (3, 11):
            sys_flags.extend([
                ("safe_path", "-P", True),
            ])
        for name, flag, value in sys_flags:
            if not isinstance(value, bool):
                if isinstance(value, int):
                    value += getattr(sys.flags, name)
            with executor_type(python_args=[flag]) as executor:
                result = executor.submit(sys_flags_get, name).result()
            if isinstance(value, bool):
                result = bool(result)
            self.assertEqual(value, result, f"sys.flags.{name}")

    @unittest.skipIf(SHARED_POOL, "shared-pool")
    def test_init_globals(self):
        executor = self.executor_type(globals={"global_var": 42})
        future1 = executor.submit(check_global_var, 42)
        future2 = executor.submit(check_global_var, 24)
        self.assertTrue(future1.result())
        self.assertFalse(future2.result())
        executor.shutdown()

    @unittest.skipIf(SHARED_POOL and WORLD_SIZE == 1, "shared-pool")
    def test_run_name(self):
        executor = self.executor_type()
        run_name = futures._core.MAIN_RUN_NAME
        future = executor.submit(check_run_name, run_name)
        self.assertTrue(future.result())

    @unittest.skipIf(SHARED_POOL and WORLD_SIZE > 2, "shared-pool")
    def test_max_workers(self):
        executor = self.executor_type(max_workers=1)
        self.assertEqual(executor.num_workers, 1)
        self.assertEqual(executor.num_workers, executor._max_workers)
        executor.shutdown()
        self.assertEqual(executor.num_workers, 0)
        self.assertEqual(executor.num_workers, executor._max_workers)

    @unittest.skipIf(SHARED_POOL and WORLD_SIZE > 2, "shared-pool")
    def test_max_workers_environ(self):
        save = os.environ.get("MPI4PY_FUTURES_MAX_WORKERS")
        os.environ["MPI4PY_FUTURES_MAX_WORKERS"] = "1"
        try:
            executor = self.executor_type()
            executor.submit(time.sleep, 0).result()
            executor.shutdown()
            executor = self.executor_type()
            self.assertEqual(executor.num_workers, 1)
            executor.shutdown()
            self.assertEqual(executor.num_workers, 0)
        finally:
            del os.environ["MPI4PY_FUTURES_MAX_WORKERS"]
            if save is not None:
                os.environ["MPI4PY_FUTURES_MAX_WORKERS"] = save

    def test_max_workers_negative(self):
        for number in (0, -1):
            with self.assertRaises(ValueError):
                self.executor_type(max_workers=number)

    def test_get_comm_workers(self):
        executor = self.executor_type()
        num_workers = executor.submit(check_comm_workers).result()
        self.assertEqual(executor.num_workers, num_workers)
        self.assertRaises(RuntimeError, check_comm_workers)

    @unittest.skipIf(SHARED_POOL, "shared-pool")
    def test_use_pkl5_kwarg(self):
        executor = self.executor_type(use_pkl5=True)
        executor.submit(time.sleep, 0).result()
        executor.shutdown()

    @unittest.skipIf(SHARED_POOL, "shared-pool")
    def test_use_pkl5_environ(self):
        save = os.environ.get("MPI4PY_FUTURES_USE_PKL5")
        try:
            for value in ("false", "true"):
                os.environ["MPI4PY_FUTURES_USE_PKL5"] = value
                executor = self.executor_type()
                executor.submit(time.sleep, 0).result()
                executor.shutdown()
            with warnings.catch_warnings(record=True) as wlist:
                warnings.simplefilter("always")
                os.environ["MPI4PY_FUTURES_USE_PKL5"] = "foobar"
                executor = self.executor_type()
                executor.submit(time.sleep, 0).result()
                executor.shutdown()
            self.assertTrue(wlist)
            msg = wlist[0].message
            self.assertIsInstance(msg, RuntimeWarning)
            self.assertIn("foobar", msg.args[0])
        finally:
            del os.environ["MPI4PY_FUTURES_USE_PKL5"]
            if save is not None:
                os.environ["MPI4PY_FUTURES_USE_PKL5"] = save

    def test_initializer(self):
        executor = self.executor_type(
            initializer=time.sleep,
            initargs=(0,),
        )
        executor.submit(time.sleep, 0).result()

    def test_initializer_bad(self):
        with self.assertRaises(TypeError):
            self.executor_type(initializer=123)

    def test_initializer_error(self):
        executor = self.executor_type(
            initializer=sleep_and_raise,
            initargs=(0.2,),
        )
        executor.submit(time.sleep, 0).cancel()
        future = executor.submit(time.sleep, 0)
        with self.assertRaises(futures.BrokenExecutor):
            executor.submit(time.sleep, 0).result()
        with self.assertRaises(futures.BrokenExecutor):
            future.result()
        with self.assertRaises(futures.BrokenExecutor):
            executor.submit(time.sleep, 0)
        self.assertEqual(executor.num_workers, 0)

    def test_initializer_error_del(self):
        executor = self.executor_type(
            initializer=sleep_and_raise,
            initargs=(0.2,),
        )
        executor.bootup()
        del executor

    def test_initializer_error_del_nowait(self):
        executor = self.executor_type(
            initializer=sleep_and_raise,
            initargs=(1.2,),
        )
        executor.bootup(wait=False)
        executor.shutdown(wait=False)
        del executor


class ProcessPoolBootupTest(ProcessPoolMixin, unittest.TestCase):
    def _prime_executor(self):
        pass

    def test_bootup(self):
        executor = self.executor_type(1)
        executor.bootup()
        executor.bootup()
        executor.shutdown()
        with self.assertRaises(RuntimeError):
            executor.bootup()

    def test_bootup_wait(self):
        executor = self.executor_type(1)
        executor.bootup(wait=True)
        executor.bootup(wait=True)
        executor.shutdown(wait=True)
        with self.assertRaises(RuntimeError):
            executor.bootup(wait=True)

    def test_bootup_nowait(self):
        executor = self.executor_type(1)
        executor.bootup(wait=False)
        executor.bootup(wait=False)
        executor.shutdown(wait=False)
        with self.assertRaises(RuntimeError):
            executor.bootup(wait=False)
        executor.shutdown(wait=True)

    def test_bootup_nowait_wait(self):
        executor = self.executor_type(1)
        executor.bootup(wait=False)
        executor.bootup(wait=True)
        executor.shutdown()
        with self.assertRaises(RuntimeError):
            executor.bootup()

    def test_bootup_shutdown_nowait(self):
        executor = self.executor_type(1)
        executor.bootup(wait=False)
        executor.shutdown(wait=False)
        worker = executor._pool
        del executor
        worker.join()


class ExecutorShutdownTestMixin:
    def test_run_after_shutdown(self):
        self.executor.shutdown()
        with self.assertRaises(RuntimeError):
            self.executor.submit(pow, 2, 5)

    def test_hang_issue12364(self):
        fs = [self.executor.submit(time.sleep, 0.01) for _ in range(50)]
        self.executor.shutdown()
        for f in fs:
            f.result()


class ProcessPoolShutdownTest(
    ProcessPoolMixin, ExecutorShutdownTestMixin, unittest.TestCase
):
    def _prime_executor(self):
        pass

    def test_shutdown(self):
        executor = self.executor_type(max_workers=1)
        self.assertIsNone(executor._pool)
        self.assertFalse(executor._shutdown)
        executor.submit(mul, 21, 2)
        executor.submit(mul, 6, 7)
        executor.submit(mul, 3, 14)
        self.assertIsNotNone(executor._pool.thread)
        self.assertFalse(executor._shutdown)
        executor.shutdown(wait=False)
        self.assertIsNotNone(executor._pool.thread)
        self.assertTrue(executor._shutdown)
        executor.shutdown(wait=True)
        self.assertIsNone(executor._pool)
        self.assertTrue(executor._shutdown)

    def test_submit_shutdown_cancel(self):
        executor = self.executor_type(max_workers=1)
        executor.bootup()
        num_workers = executor.num_workers
        for _ in range(num_workers * 100):
            executor.submit(time.sleep, 0.1)
        fut = executor.submit(time.sleep, 0)
        executor.shutdown(wait=False, cancel_futures=False)
        self.assertFalse(fut.cancelled())
        executor.shutdown(wait=True, cancel_futures=True)
        self.assertTrue(fut.cancelled())

    def test_submit_shutdown_cancel_wait(self):
        executor = self.executor_type(max_workers=1)
        executor.bootup()
        num_workers = executor.num_workers
        fut1 = executor.submit(time.sleep, 0.1)
        for _ in range(num_workers * 100):
            executor.submit(time.sleep, 0.1)
        fut2 = executor.submit(time.sleep, 0)
        fut3 = executor.submit(time.sleep, 0)
        time.sleep(0.2)
        executor.shutdown(wait=False, cancel_futures=True)
        _done, not_done = futures.wait({fut1, fut2, fut3})
        self.assertEqual(len(not_done), 0)
        self.assertFalse(fut1.cancelled())
        self.assertTrue(fut2.cancelled())
        self.assertTrue(fut3.cancelled())
        executor.shutdown(wait=True, cancel_futures=True)

    def test_shutdown_cancel(self):
        executor = self.executor_type(max_workers=1)
        executor.bootup()
        executor._pool.cancel()
        executor.shutdown(wait=False, cancel_futures=False)
        executor.shutdown(wait=False, cancel_futures=False)
        executor.shutdown(wait=False, cancel_futures=True)
        executor.shutdown(wait=False, cancel_futures=True)
        executor.shutdown(wait=True, cancel_futures=True)
        executor.shutdown(wait=True, cancel_futures=True)

    def test_init_bootup_shutdown(self):
        executor = self.executor_type(max_workers=1)
        self.assertIsNone(executor._pool)
        self.assertFalse(executor._shutdown)
        executor.bootup()
        self.assertTrue(executor._pool.event.is_set())
        self.assertFalse(executor._shutdown)
        executor.shutdown()
        self.assertIsNone(executor._pool)
        self.assertTrue(executor._shutdown)

    def test_context_manager_shutdown(self):
        with self.executor_type(max_workers=1) as e:
            self.assertEqual(
                list(e.map(abs, range(-5, 5))), [5, 4, 3, 2, 1, 0, 1, 2, 3, 4]
            )
            threads = [e._pool.thread]
            queues = [e._pool.queue]
            events = [e._pool.event]

        for t in threads:
            t.join()
        for q in queues:
            with self.assertRaises(LookupError):
                q.pop()
        for e in events:
            self.assertTrue(e.is_set())

    def test_del_shutdown(self):
        executor = self.executor_type(max_workers=1)
        list(executor.map(abs, range(-5, 5)))
        threads = [executor._pool.thread]
        queues = [executor._pool.queue]
        events = [executor._pool.event]
        if hasattr(sys, "pypy_version_info"):
            executor.shutdown(False)
        else:
            del executor

        for t in threads:
            t.join()
        for q in queues:
            with self.assertRaises(LookupError):
                q.pop()
        for e in events:
            self.assertTrue(e.is_set())


class WaitTestMixin:
    def test_first_completed(self):
        future1 = self.executor.submit(mul, 21, 2)
        future2 = self.executor.submit(time.sleep, 0.5)

        done, not_done = futures.wait(
            [CANCELLED_FUTURE, future1, future2],
            return_when=futures.FIRST_COMPLETED,
        )

        self.assertEqual({future1}, done)
        self.assertEqual({CANCELLED_FUTURE, future2}, not_done)

    def test_first_completed_some_already_completed(self):
        future1 = self.executor.submit(time.sleep, 0.5)

        finished, pending = futures.wait(
            [CANCELLED_AND_NOTIFIED_FUTURE, SUCCESSFUL_FUTURE, future1],
            return_when=futures.FIRST_COMPLETED,
        )

        self.assertEqual(
            {CANCELLED_AND_NOTIFIED_FUTURE, SUCCESSFUL_FUTURE}, finished
        )
        self.assertEqual({future1}, pending)

    def test_first_exception(self):
        future1 = self.executor.submit(mul, 2, 21)
        future2 = self.executor.submit(sleep_and_raise, 0.25)
        future3 = self.executor.submit(time.sleep, 0.5)

        finished, pending = futures.wait(
            [future1, future2, future3], return_when=futures.FIRST_EXCEPTION
        )

        self.assertEqual({future1, future2}, finished)
        self.assertEqual({future3}, pending)

    def test_first_exception_some_already_complete(self):
        future1 = self.executor.submit(divmod, 21, 0)
        future2 = self.executor.submit(time.sleep, 0.5)

        finished, pending = futures.wait(
            [
                SUCCESSFUL_FUTURE,
                CANCELLED_FUTURE,
                CANCELLED_AND_NOTIFIED_FUTURE,
                future1,
                future2,
            ],
            return_when=futures.FIRST_EXCEPTION,
        )

        self.assertEqual(
            {SUCCESSFUL_FUTURE, CANCELLED_AND_NOTIFIED_FUTURE, future1},
            finished,
        )
        self.assertEqual({CANCELLED_FUTURE, future2}, pending)

    def test_first_exception_one_already_failed(self):
        future1 = self.executor.submit(time.sleep, 0.25)

        finished, pending = futures.wait(
            [EXCEPTION_FUTURE, future1], return_when=futures.FIRST_EXCEPTION
        )

        self.assertEqual({EXCEPTION_FUTURE}, finished)
        self.assertEqual({future1}, pending)

    def test_all_completed(self):
        future1 = self.executor.submit(divmod, 2, 0)
        future2 = self.executor.submit(mul, 2, 21)

        finished, pending = futures.wait(
            [
                SUCCESSFUL_FUTURE,
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                future1,
                future2,
            ],
            return_when=futures.ALL_COMPLETED,
        )

        self.assertEqual(
            {
                SUCCESSFUL_FUTURE,
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                future1,
                future2,
            },
            finished,
        )
        self.assertEqual(set(), pending)

    def test_timeout(self):
        future1 = self.executor.submit(mul, 6, 7)
        future2 = self.executor.submit(time.sleep, 0.75)

        finished, pending = futures.wait(
            [
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                SUCCESSFUL_FUTURE,
                future1,
                future2,
            ],
            timeout=0.5,
            return_when=futures.ALL_COMPLETED,
        )

        self.assertEqual(
            {
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                SUCCESSFUL_FUTURE,
                future1,
            },
            finished,
        )
        self.assertEqual({future2}, pending)


class ProcessPoolWaitTest(ProcessPoolMixin, WaitTestMixin, unittest.TestCase):
    pass


class AsCompletedTestMixin:
    def test_no_timeout(self):
        future1 = self.executor.submit(mul, 2, 21)
        future2 = self.executor.submit(mul, 7, 6)

        completed = set(
            futures.as_completed([
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                SUCCESSFUL_FUTURE,
                future1,
                future2,
            ])
        )
        self.assertEqual(
            {
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                SUCCESSFUL_FUTURE,
                future1,
                future2,
            },
            completed,
        )

    def test_zero_timeout(self):
        future1 = self.executor.submit(time.sleep, 0.5)
        completed_futures = set()
        try:
            completed_futures.update(
                futures.as_completed(
                    [
                        CANCELLED_AND_NOTIFIED_FUTURE,
                        EXCEPTION_FUTURE,
                        SUCCESSFUL_FUTURE,
                        future1,
                    ],
                    timeout=0,
                )
            )
        except futures.TimeoutError:
            pass

        self.assertEqual(
            {
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                SUCCESSFUL_FUTURE,
            },
            completed_futures,
        )

    def test_nonzero_timeout(self):
        future1 = self.executor.submit(time.sleep, 0.0)
        self.executor.submit(time.sleep, 0.5)
        completed_futures = set()
        try:
            completed_futures.update(
                futures.as_completed(
                    [
                        CANCELLED_AND_NOTIFIED_FUTURE,
                        EXCEPTION_FUTURE,
                        SUCCESSFUL_FUTURE,
                        future1,
                    ],
                    timeout=0.2,
                )
            )
        except futures.TimeoutError:
            pass

        self.assertEqual(
            {
                CANCELLED_AND_NOTIFIED_FUTURE,
                EXCEPTION_FUTURE,
                SUCCESSFUL_FUTURE,
                future1,
            },
            completed_futures,
        )

    def test_duplicate_futures(self):
        # Issue 20367. Duplicate futures should not raise exceptions or give
        # duplicate responses.
        future1 = self.executor.submit(time.sleep, 0.1)
        completed = list(futures.as_completed([future1, future1]))
        self.assertEqual(len(completed), 1)


class ProcessPoolAsCompletedTest(
    ProcessPoolMixin, AsCompletedTestMixin, unittest.TestCase
):
    pass


class ExecutorTestMixin:
    def test_submit(self):
        future = self.executor.submit(pow, 2, 8)
        self.assertEqual(256, future.result())

    def test_submit_keyword(self):
        future = self.executor.submit(mul, 2, y=8)
        self.assertEqual(16, future.result())
        future = self.executor.submit(mul, x=2, y=8)
        self.assertEqual(16, future.result())

    def test_submit_cancel(self):
        fs = []
        num_workers = self.executor.num_workers
        for _ in range(num_workers * 100):
            f = self.executor.submit(time.sleep, 0.1)
            fs.append(f)
        future = self.executor.submit(time.sleep, 0)
        future.cancel()
        self.assertTrue(future.cancelled())
        for f in fs:
            f.cancel()

    def test_map(self):
        self.assertEqual(
            list(self.executor.map(pow, range(10), range(10))),
            list(map(pow, range(10), range(10))),
        )

    def test_starmap(self):
        sequence = [(a, a) for a in range(10)]
        self.assertEqual(
            list(self.executor.starmap(pow, sequence)),
            list(map(pow, range(10), range(10))),
        )
        self.assertEqual(
            list(self.executor.starmap(pow, iter(sequence))),
            list(map(pow, range(10), range(10))),
        )

    def test_map_exception(self):
        i = self.executor.map(divmod, [1, 1, 1, 1], [2, 3, 0, 5])
        self.assertEqual(next(i), (0, 1))
        self.assertEqual(next(i), (0, 1))
        with self.assertRaises(ZeroDivisionError):
            next(i)

    def test_map_timeout(self):
        results = []
        try:
            for i in self.executor.map(time.sleep, [0, 0, 1], timeout=0.25):
                results.append(i)
        except futures.TimeoutError:
            pass
        else:
            self.fail("expected TimeoutError")

        self.assertEqual([None, None], results)

    def test_map_timeout_one(self):
        results = []
        for i in self.executor.map(time.sleep, [0, 0, 0], timeout=1):
            results.append(i)
        self.assertEqual([None, None, None], results)


class ProcessPoolExecutorTest(
    ProcessPoolMixin, ExecutorTestMixin, unittest.TestCase
):
    def test_map_chunksize(self):
        ref = list(map(pow, range(40), range(40)))
        self.assertEqual(
            list(self.executor.map(pow, range(40), range(40), chunksize=6)),
            ref,
        )
        self.assertEqual(
            list(self.executor.map(pow, range(40), range(40), chunksize=50)),
            ref,
        )
        self.assertEqual(
            list(self.executor.map(pow, range(40), range(40), chunksize=40)),
            ref,
        )
        with self.assertRaises(ValueError):
            list(self.executor.map(pow, range(40), range(40), chunksize=-1))

    def test_starmap_chunksize(self):
        ref = list(map(pow, range(40), range(40)))
        sequence = [(a, a) for a in range(40)]
        self.assertEqual(
            list(self.executor.starmap(pow, sequence, chunksize=6)), ref
        )
        self.assertEqual(
            list(self.executor.starmap(pow, sequence, chunksize=50)), ref
        )
        self.assertEqual(
            list(self.executor.starmap(pow, sequence, chunksize=40)), ref
        )
        self.assertEqual(
            list(self.executor.starmap(pow, iter(sequence), chunksize=6)), ref
        )
        self.assertEqual(
            list(self.executor.starmap(pow, iter(sequence), chunksize=50)), ref
        )
        self.assertEqual(
            list(self.executor.starmap(pow, iter(sequence), chunksize=40)), ref
        )

        with self.assertRaises(ValueError):
            list(self.executor.starmap(pow, sequence, chunksize=-1))

    def test_map_buffersize(self):
        ref = list(map(pow, range(10), range(10)))
        for bs in (1, 2, 9, 10, 11):
            for cs in (1, 2, 9, 10, 11):
                self.assertEqual(
                    list(
                        self.executor.map(
                            pow,
                            range(10),
                            range(10),
                            chunksize=cs,
                            buffersize=bs,
                        )
                    ),
                    ref,
                )
        with self.assertRaises(TypeError):
            self.executor.map(
                pow,
                range(10),
                range(10),
                buffersize=str(1),
            )
        with self.assertRaises(ValueError):
            self.executor.map(
                pow,
                range(10),
                range(10),
                buffersize=-1,
            )
        with self.assertRaises(ValueError):
            self.executor.map(
                pow,
                range(10),
                range(10),
                buffersize=1,
                unordered=True,
            )

    def test_map_unordered(self):
        map_unordered = functools.partial(self.executor.map, unordered=True)
        self.assertEqual(
            set(map_unordered(pow, range(10), range(10))),
            set(map(pow, range(10), range(10))),
        )

    def test_map_unordered_timeout(self):
        map_unordered = functools.partial(self.executor.map, unordered=True)
        num_workers = self.executor.num_workers
        results = []
        try:
            args = [1] + [0] * (num_workers - 1)
            for i in map_unordered(time.sleep, args, timeout=0.25):
                results.append(i)
        except futures.TimeoutError:
            pass
        else:
            self.fail("expected TimeoutError")

        self.assertEqual([None] * (num_workers - 1), results)

    def test_map_unordered_timeout_one(self):
        map_unordered = functools.partial(self.executor.map, unordered=True)
        results = []
        for i in map_unordered(time.sleep, [0, 0, 0], timeout=1):
            results.append(i)
        self.assertEqual([None, None, None], results)

    def test_map_unordered_exception(self):
        map_unordered = functools.partial(self.executor.map, unordered=True)
        i = map_unordered(divmod, [1, 1, 1, 1], [2, 3, 0, 5])
        try:
            self.assertEqual(next(i), (0, 1))
        except ZeroDivisionError:
            return

    def test_map_unordered_chunksize(self):
        map_unordered = functools.partial(self.executor.map, unordered=True)
        ref = set(map(pow, range(40), range(40)))
        self.assertEqual(
            set(map_unordered(pow, range(40), range(40), chunksize=6)), ref
        )
        self.assertEqual(
            set(map_unordered(pow, range(40), range(40), chunksize=50)), ref
        )
        self.assertEqual(
            set(map_unordered(pow, range(40), range(40), chunksize=40)), ref
        )
        with self.assertRaises(ValueError):
            set(map_unordered(pow, range(40), range(40), chunksize=-1))


class ProcessPoolSubmitTest(unittest.TestCase):
    @unittest.skipIf(MPI.get_vendor()[0] == "Microsoft MPI", "msmpi")
    def test_multiple_executors(self):
        executor1 = futures.MPIPoolExecutor(1).bootup(wait=True)
        executor2 = futures.MPIPoolExecutor(1).bootup(wait=True)
        executor3 = futures.MPIPoolExecutor(1).bootup(wait=True)
        fs1 = [executor1.submit(abs, i) for i in range(100, 200)]
        fs2 = [executor2.submit(abs, i) for i in range(200, 300)]
        fs3 = [executor3.submit(abs, i) for i in range(300, 400)]
        futures.wait(fs3 + fs2 + fs1)
        for i, f in enumerate(fs1):
            self.assertEqual(f.result(), i + 100)
        for i, f in enumerate(fs2):
            self.assertEqual(f.result(), i + 200)
        for i, f in enumerate(fs3):
            self.assertEqual(f.result(), i + 300)
        executor1 = executor2 = executor3 = None

    def test_mpi_serialized_support(self):
        futures._core.setup_mpi_threads()
        serialized = futures._core.serialized
        lock_save = serialized.lock
        try:
            if lock_save is None:
                serialized.lock = threading.Lock()
                executor = futures.MPIPoolExecutor(1).bootup()
                executor.submit(abs, 0).result()
                executor.shutdown()
                serialized.lock = lock_save
            else:
                serialized.lock = None
                with lock_save:
                    executor = futures.MPIPoolExecutor(1).bootup()
                    executor.submit(abs, 0).result()
                    executor.shutdown()
                serialized.lock = lock_save
        finally:
            serialized.lock = lock_save

    def test_shared_executors(self):
        if not SHARED_POOL:
            return
        executors = [futures.MPIPoolExecutor() for _ in range(16)]
        fs = []
        for i in range(128):
            fs.extend(
                e.submit(abs, i * 16 + j) for j, e in enumerate(executors)
            )
        self.assertEqual(sorted(f.result() for f in fs), list(range(16 * 128)))
        world_size = MPI.COMM_WORLD.Get_size()
        num_workers = max(1, world_size - 1)
        for e in executors:
            self.assertEqual(e.num_workers, num_workers)
        del e, executors


def inout(arg):
    return arg


class GoodPickle:
    def __init__(self, value=0):
        self.value = value
        self.pickled = False
        self.unpickled = False

    def __getstate__(self):
        self.pickled = True
        return (self.value,)

    def __setstate__(self, state):
        self.unpickled = True
        self.value = state[0]


class BadPickle:
    def __init__(self):
        self.pickled = False

    def __getstate__(self):
        self.pickled = True
        _ = 1 / 0

    def __setstate__(self, state):
        pass


class BadUnpickle:
    def __init__(self):
        self.pickled = False

    def __getstate__(self):
        self.pickled = True
        return (None,)

    def __setstate__(self, state):
        if state[0] is not None:
            raise ValueError
        _ = 1 / 0


@unittest.skipIf(SHARED_POOL and WORLD_SIZE == 1, "shared-pool")
class ProcessPoolPickleTest(unittest.TestCase):
    def setUp(self):
        self.executor = futures.MPIPoolExecutor(1)

    def tearDown(self):
        self.executor.shutdown()

    def test_good_pickle(self):
        o = GoodPickle(42)
        r = self.executor.submit(inout, o).result()
        self.assertEqual(o.value, r.value)
        self.assertTrue(o.pickled)
        self.assertTrue(r.unpickled)

        r = self.executor.submit(GoodPickle, 77).result()
        self.assertEqual(r.value, 77)
        self.assertTrue(r.unpickled)

    def test_bad_pickle(self):
        o = BadPickle()
        self.assertFalse(o.pickled)
        f = self.executor.submit(inout, o)
        with self.assertRaises(ZeroDivisionError):
            f.result()
        self.assertTrue(o.pickled)

        f = self.executor.submit(BadPickle)
        with self.assertRaises(ZeroDivisionError):
            f.result()

        f = self.executor.submit(abs, 42)
        self.assertEqual(f.result(), 42)

    def test_bad_unpickle(self):
        o = BadUnpickle()
        self.assertFalse(o.pickled)
        f = self.executor.submit(inout, o)
        with self.assertRaises(ZeroDivisionError):
            f.result()
        self.assertTrue(o.pickled)

        f = self.executor.submit(BadUnpickle)
        with self.assertRaises(ZeroDivisionError):
            f.result()

        f = self.executor.submit(abs, 42)
        self.assertEqual(f.result(), 42)

    def test_exc_pickle(self):
        o = BadPickle()
        f = self.executor.submit(inout, o)
        exc = f.exception()
        self.assertIsInstance(exc, ZeroDivisionError)
        cause = exc.__cause__
        self.assertIsNone(cause)

    def test_exc_unpickle(self):
        o = BadUnpickle()
        f = self.executor.submit(inout, o)
        exc = f.exception()
        self.assertIsInstance(exc, ZeroDivisionError)
        cause = exc.__cause__
        self.assertIsInstance(cause, futures._core.RemoteTraceback)


class MPICommExecutorTest(unittest.TestCase):
    MPICommExecutor = futures.MPICommExecutor

    def test_default(self):
        with self.MPICommExecutor() as executor:
            if executor is not None:
                executor.bootup()
                future1 = executor.submit(time.sleep, 0)
                future2 = executor.submit(time.sleep, 0)
                executor.shutdown()
                self.assertIsNone(future1.result())
                self.assertIsNone(future2.result())

    def test_self(self):
        with self.MPICommExecutor(MPI.COMM_SELF) as executor:
            future = executor.submit(time.sleep, 0)
            self.assertIsNone(future.result())
            self.assertIsNone(future.exception())

            future = executor.submit(sleep_and_raise, 0)
            with self.assertRaises(RuntimeError):
                future.result()
            self.assertEqual(RuntimeError, type(future.exception()))

            list(executor.map(time.sleep, [0, 0]))
            list(executor.map(time.sleep, [0, 0], timeout=1))
            iterator = executor.map(time.sleep, [0.2, 0], timeout=0)
            with self.assertRaises(futures.TimeoutError):
                list(iterator)

    def test_args(self):
        with self.MPICommExecutor(MPI.COMM_SELF) as executor:
            self.assertIsNotNone(executor)
        with self.MPICommExecutor(MPI.COMM_SELF, 0) as executor:
            self.assertIsNotNone(executor)

    def test_kwargs(self):
        with self.MPICommExecutor(comm=MPI.COMM_SELF) as executor:
            self.assertIsNotNone(executor)
        with self.MPICommExecutor(comm=MPI.COMM_SELF, root=0) as executor:
            self.assertIsNotNone(executor)

    @unittest.skipIf(SHARED_POOL, "shared-pool")
    def test_arg_root(self):
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        for root in range(comm.Get_size()):
            with self.MPICommExecutor(comm, root) as executor:
                if rank == root:
                    self.assertIsNotNone(executor)
                else:
                    self.assertIsNone(executor)
            with self.MPICommExecutor(root=root) as executor:
                if rank == root:
                    self.assertIsNotNone(executor)
                else:
                    self.assertIsNone(executor)

    @unittest.skipIf(SHARED_POOL, "shared-pool")
    def test_arg_bad_root(self):
        size = MPI.COMM_WORLD.Get_size()
        with self.assertRaises(ValueError):
            self.MPICommExecutor(root=-size)
        with self.assertRaises(ValueError):
            self.MPICommExecutor(root=-1)
        with self.assertRaises(ValueError):
            self.MPICommExecutor(root=+size)

    @unittest.skipIf(SHARED_POOL, "shared-pool")
    def test_arg_bad_comm(self):
        if MPI.COMM_WORLD.Get_size() == 1:
            return
        intercomm, intracomm = futures._core.comm_split(MPI.COMM_WORLD, 0)
        try:
            with self.assertRaises(ValueError):
                self.MPICommExecutor(intercomm)
        finally:
            intercomm.Free()
            if intracomm:
                intracomm.Free()

    def test_with_bad(self):
        mpicommexecutor = self.MPICommExecutor(MPI.COMM_SELF)
        with mpicommexecutor:
            try:
                with mpicommexecutor:
                    pass
            except RuntimeError:
                pass
            else:
                self.fail("expected RuntimeError")

    def test_initializer(self):
        mpicommexecutor = self.MPICommExecutor(
            initializer=time.sleep,
            initargs=(0,),
        )
        with mpicommexecutor as executor:
            if executor is not None:
                executor.bootup()
                del executor
        with mpicommexecutor as executor:
            if executor is not None:
                executor.submit(time.sleep, 0).result()

    def test_initializer_error(self):
        mpicommexecutor = self.MPICommExecutor(
            initializer=sleep_and_raise,
            initargs=(0.2,),
        )
        with mpicommexecutor as executor:
            if executor is not None:
                with contextlib.suppress(futures.BrokenExecutor):
                    executor.submit(time.sleep, 0).cancel()
                future = executor.submit(time.sleep, 0)
                with self.assertRaises(futures.BrokenExecutor):
                    executor.submit(time.sleep, 0).result()
                with self.assertRaises(futures.BrokenExecutor):
                    future.result()

    def test_initializer_error_del(self):
        mpicommexecutor = self.MPICommExecutor(
            initializer=sleep_and_raise,
            initargs=(0.2,),
        )
        with mpicommexecutor as executor:
            if executor is not None:
                executor.bootup()
                del executor

    def test_initializer_error_del_nowait(self):
        mpicommexecutor = self.MPICommExecutor(
            initializer=sleep_and_raise,
            initargs=(0.2,),
        )
        with mpicommexecutor as executor:
            if executor is not None:
                executor.bootup(wait=False)
                executor.shutdown(wait=False)
                del executor

    def test_get_comm_workers(self):
        for _comm in (MPI.COMM_SELF, MPI.COMM_WORLD):
            with self.MPICommExecutor(MPI.COMM_SELF) as executor:
                num_workers = executor.submit(check_comm_workers).result()
                self.assertEqual(executor.num_workers, num_workers)
        self.assertRaises(RuntimeError, check_comm_workers)


class ThreadPoolMixin(ExecutorMixin):
    executor_type = futures.ThreadPoolExecutor


class ThreadPoolTest(
    ThreadPoolMixin,
    ExecutorTestMixin,
    ExecutorShutdownTestMixin,
    unittest.TestCase,
):
    pass


from mpi4py.futures.aplus import ThenableFuture  # noqa: E402


class ThenTest(unittest.TestCase):
    assert_ = unittest.TestCase.assertTrue

    def test_cancel_base(self):
        base_f = ThenableFuture()
        new_f = base_f.then()

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.cancel()
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(base_f.cancelled())
        self.assertTrue(new_f.cancelled())

    def test_cancel_new(self):
        base_f = ThenableFuture()
        new_f = base_f.then()

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        new_f.cancel()
        self.assertTrue(not base_f.done())
        self.assertTrue(new_f.done())

        base_f.set_result(1)
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.cancelled())

    def test_then_multiple(self):
        base_f = ThenableFuture()
        new_f1 = base_f.then()
        new_f2 = base_f.then()
        new_f3 = base_f.then()

        self.assertIsNot(base_f, new_f1)
        self.assertIsNot(base_f, new_f2)
        self.assertIsNot(base_f, new_f3)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f1.done())
        self.assertTrue(not new_f2.done())
        self.assertTrue(not new_f3.done())

        base_f.set_result("done")
        self.assertTrue(base_f.done())
        self.assertTrue(new_f1.done())
        self.assertTrue(new_f2.done())
        self.assertTrue(new_f3.done())

        self.assertTrue(not new_f1.exception())
        self.assertTrue(not new_f2.exception())
        self.assertTrue(not new_f3.exception())
        self.assertEqual(new_f1.result(), "done")
        self.assertEqual(new_f2.result(), "done")
        self.assertEqual(new_f3.result(), "done")

    def test_no_callbacks_and_success(self):
        base_f = ThenableFuture()
        new_f = base_f.then()

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_result("done")
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(not new_f.exception())
        self.assertEqual(new_f.result(), "done")

    def test_no_callbacks_and_failure(self):
        class MyException(Exception):
            pass

        base_f = ThenableFuture()
        new_f = base_f.then()

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_exception(MyException("sad"))
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(new_f.exception())
        with self.assertRaises(MyException) as catcher:
            new_f.result()
        self.assertEqual(catcher.exception.args[0], "sad")

    def test_success_callback_and_success(self):
        base_f = ThenableFuture()
        new_f = base_f.then(lambda result: result + " manipulated")

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_result("done")
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(not new_f.exception())
        self.assertEqual(new_f.result(), "done manipulated")

    def test_err_callback_and_failure_repackage(self):
        class MyException(Exception):
            pass

        class MyRepackagedException(Exception):
            pass

        class NotMatched(Exception):
            pass

        def on_failure(ex):
            if isinstance(ex, MyException):
                return MyRepackagedException(ex.args[0] + " repackaged")
            else:
                return NotMatched("?")

        base_f = ThenableFuture()
        new_f = base_f.then(None, on_failure)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_exception(MyException("sad"))
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(new_f.exception())
        with self.assertRaises(MyRepackagedException) as catcher:
            new_f.result()
        self.assertEqual(catcher.exception.args[0], "sad repackaged")

    def test_err_callback_and_failure_raised(self):
        class MyException(Exception):
            pass

        class MyRepackagedException(Exception):
            pass

        def raise_something_else(ex):
            raise MyRepackagedException(ex.args[0] + " repackaged")

        base_f = ThenableFuture()
        new_f = base_f.then(None, raise_something_else)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_exception(MyException("sad"))
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(new_f.exception())
        with self.assertRaises(MyRepackagedException) as catcher:
            new_f.result()
        self.assertEqual(catcher.exception.args[0], "sad repackaged")

    def test_err_callback_convert_to_success(self):
        class MyException(Exception):
            pass

        class NotMatched(Exception):
            pass

        def on_failure(ex):
            if isinstance(ex, MyException):
                return ex.args[0] + " repackaged"
            else:
                return NotMatched("?")

        base_f = ThenableFuture()
        new_f = base_f.catch(on_failure)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_exception(MyException("sad"))
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(not new_f.exception())
        self.assertEqual(new_f.result(), "sad repackaged")

    def test_err_catch_ignore(self):
        base_f = ThenableFuture()
        new_f = base_f.catch()

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_exception(Exception("sad"))
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertIsNone(new_f.exception())
        self.assertIsNone(new_f.result())

    def test_success_callback_and_failure_raised(self):
        class MyException(Exception):
            pass

        def raise_something_else(value):
            raise MyException(value + " repackaged")

        base_f = ThenableFuture()
        new_f = base_f.then(raise_something_else)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_result("sad")
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(new_f.exception())
        with self.assertRaises(MyException) as catcher:
            new_f.result()
        assert catcher.exception.args[0] == "sad repackaged"

    def test_chained_success_callback_and_success(self):
        def transform(value):
            f = ThenableFuture()
            if value < 5:
                f.set_result(transform(value + 1))
            else:
                f.set_result(value)
            return f

        base_f = ThenableFuture()
        new_f = base_f.then(transform)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_result(1)
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(not new_f.exception())
        self.assertEqual(new_f.result(), 5)

    def test_chained_failure_callback_and_success(self):
        def transform(exc):
            self.assertIsInstance(exc, RuntimeError)
            f = ThenableFuture()
            f.set_result(5)
            return f

        base_f = ThenableFuture()
        new_f = base_f.catch(transform)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_exception(RuntimeError())
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(not new_f.exception())
        self.assertEqual(new_f.result(), 5)

    def test_detect_cycle_chain(self):
        f1 = ThenableFuture()
        f2 = ThenableFuture()
        chain = [f1, f2, f1]

        def transform(a):
            try:
                f = chain.pop(0)
                r = transform(a)
                f.__init__()
                f.set_result(r)
            except IndexError:
                return 42
            else:
                return f

        base_f = ThenableFuture()
        new_f = base_f.then(transform)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_result(1)
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(new_f.exception())
        with self.assertRaises(RuntimeError) as catcher:
            new_f.result()
        self.assertIn(
            "chain cycle detected",
            catcher.exception.args[0],
        )

    def test_detect_self_chain(self):
        base_f = ThenableFuture()
        new_f = base_f.then(lambda _: new_f)

        self.assertIsNot(base_f, new_f)
        self.assertTrue(not base_f.done())
        self.assertTrue(not new_f.done())

        base_f.set_result(1)
        self.assertTrue(base_f.done())
        self.assertTrue(new_f.done())

        self.assertTrue(new_f.exception())
        with self.assertRaises(RuntimeError) as catcher:
            new_f.result()
        self.assertIn(
            "chain cycle detected",
            catcher.exception.args[0],
        )


class CollectTest(unittest.TestCase):
    def test_empty(self):
        future = futures.collect([])
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertTrue(future.done())
        self.assertEqual(future.result(), [])

    def test_item_success(self):
        fs = [futures.Future() for _ in range(5)]
        future = futures.collect(fs)
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertFalse(future.done())
        for i in range(5):
            fs[i].set_result(i)
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertTrue(future.done())
        self.assertEqual(future.result(), list(range(5)))

    def test_item_failure(self):
        fs = [futures.Future() for _ in range(5)]
        future = futures.collect(fs)
        for i in range(2, 4):
            fs[i].set_result(i)
        fs[-1].set_exception(RuntimeError())
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertTrue(future.done())
        self.assertIsInstance(future.exception(), RuntimeError)
        for i in range(2):
            self.assertTrue(fs[i].cancelled())
        for i in range(2, 4):
            self.assertFalse(fs[i].cancelled())
        self.assertFalse(fs[-1].cancelled())

    def test_item_done(self):
        fs = [futures.Future() for _ in range(5)]
        for i in range(5):
            fs[i].set_result(i)
        future = futures.collect(fs)
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertTrue(future.done())
        self.assertEqual(future.result(), list(range(5)))

    def test_item_cancel(self):
        fs = [futures.Future() for _ in range(5)]
        future = futures.collect(fs)
        for i in range(2, 4):
            fs[i].set_result(i)
        fs[-1].cancel()
        self.assertTrue(future.cancelled())
        self.assertFalse(future.running())
        self.assertTrue(future.done())
        for i in range(2):
            self.assertTrue(fs[i].cancelled())
        for i in range(2, 4):
            self.assertFalse(fs[i].cancelled())
        self.assertTrue(fs[-1].cancelled())

    def test_cancel(self):
        fs = [futures.Future() for _ in range(5)]
        future = futures.collect(fs)
        future.cancel()
        for f in fs:
            self.assertTrue(f.cancelled())

    def test_cancel_pending(self):
        class MyFuture(futures.Future):
            def cancel(self):
                pass

        fs = [MyFuture() for _ in range(5)]
        future = futures.collect(fs)
        self.assertIs(type(future), MyFuture)
        super(MyFuture, future).cancel()
        for f in fs:
            self.assertFalse(f.cancelled())
            f.set_result(None)


class ComposeTest(unittest.TestCase):
    def test_result(self):
        base = futures.Future()
        future = futures.compose(base)
        self.assertIs(type(future), type(base))
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertFalse(future.done())
        base.set_result(42)
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertTrue(future.done())
        self.assertEqual(future.result(), 42)

    def test_except(self):
        base = futures.Future()
        future = futures.compose(base)
        self.assertIs(type(future), type(base))
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertFalse(future.done())
        base.set_exception(RuntimeError(42))
        self.assertFalse(future.cancelled())
        self.assertFalse(future.running())
        self.assertTrue(future.done())
        self.assertIs(type(future.exception()), RuntimeError)
        self.assertEqual(future.exception().args, (42,))

    def test_cancel_new(self):
        base = futures.Future()
        future = futures.compose(base)
        base.cancel()
        self.assertTrue(future.cancelled())

    def test_cancel_old(self):
        base = futures.Future()
        future = futures.compose(base)
        future.cancel()
        self.assertTrue(base.cancelled())

    def test_result_hook(self):
        base = futures.Future()
        future = futures.compose(base, int)
        base.set_result("42")
        self.assertEqual(future.result(), 42)

    def test_result_hook_failure(self):
        base = futures.Future()
        future = futures.compose(base, resulthook=lambda _: 1 / 0)
        base.set_result(42)
        self.assertIs(type(future.exception()), ZeroDivisionError)

    def test_except_hook(self):
        base = futures.Future()
        future = futures.compose(base, excepthook=lambda exc: exc.args[0])
        base.set_exception(RuntimeError(42))
        self.assertEqual(future.result(), 42)

    def test_except_hook_except(self):
        base = futures.Future()
        future = futures.compose(
            base, excepthook=lambda exc: RuntimeError(exc.args[0])
        )
        base.set_exception(ValueError(42))
        self.assertIs(type(future.exception()), RuntimeError)
        self.assertEqual(future.exception().args, (42,))

    def test_except_hook_failure(self):
        base = futures.Future()
        future = futures.compose(base, excepthook=lambda _: 1 / 0)
        base.set_exception(ValueError(42))
        self.assertIs(type(future.exception()), ZeroDivisionError)


SKIP_POOL_TEST = testutil.disable_mpi_spawn()

if SHARED_POOL:
    del ProcessPoolInitTest.test_init_sys_flags
    del ProcessPoolInitTest.test_init_globals
    del ProcessPoolInitTest.test_use_pkl5_kwarg
    del ProcessPoolInitTest.test_use_pkl5_environ
    if WORLD_SIZE == 1:
        del ProcessPoolInitTest.test_run_name
    if WORLD_SIZE > 2:
        del ProcessPoolInitTest.test_max_workers
        del ProcessPoolInitTest.test_max_workers_environ
    if WORLD_SIZE == 1:
        del ProcessPoolPickleTest
    del MPICommExecutorTest.test_arg_root
    del MPICommExecutorTest.test_arg_bad_root
    del MPICommExecutorTest.test_arg_bad_comm
elif WORLD_SIZE > 1 or SKIP_POOL_TEST:
    del ProcessPoolInitTest
    del ProcessPoolBootupTest
    del ProcessPoolShutdownTest
    del ProcessPoolWaitTest
    del ProcessPoolAsCompletedTest
    del ProcessPoolExecutorTest
    del ProcessPoolSubmitTest
    del ProcessPoolPickleTest
if not SHARED_POOL:
    del SharedPoolInitTest


if __name__ == "__main__":
    unittest.main()



import ctypes
import sys
from enum import IntEnum

if hasattr(sys, "pypy_version_info"):
    raise ImportError("unsupported on PyPy")


class DLPackVersion(ctypes.Structure):
    #
    _fields_ = [
        ("major", ctypes.c_uint32),
        ("minor", ctypes.c_uint32),
    ]


class DLDeviceType(IntEnum):
    #
    kDLCPU = 1
    kDLCUDA = 2
    kDLCUDAHost = 3
    kDLOpenCL = 4
    kDLVulkan = 7
    kDLMetal = 8
    kDLVPI = 9
    kDLROCM = 10
    kDLROCMHost = 11
    kDLExtDev = 12
    kDLCUDAManaged = 13
    kDLOneAPI = 14
    kDLWebGPU = 15
    kDLHexagon = 16


class DLDevice(ctypes.Structure):
    #
    _fields_ = [
        ("device_type", ctypes.c_uint),
        ("device_id", ctypes.c_int32),
    ]


class DLDataTypeCode(IntEnum):
    #
    kDLInt = 0
    kDLUInt = 1
    kDLFloat = 2
    kDLOpaqueHandle = 3
    kDLBfloat = 4
    kDLComplex = 5
    kDLBool = 6


class DLDataType(ctypes.Structure):
    #
    _fields_ = [
        ("code", ctypes.c_uint8),
        ("bits", ctypes.c_uint8),
        ("lanes", ctypes.c_uint16),
    ]


class DLTensor(ctypes.Structure):
    #
    _fields_ = [
        ("data", ctypes.c_void_p),
        ("device", DLDevice),
        ("ndim", ctypes.c_int32),
        ("dtype", DLDataType),
        ("shape", ctypes.POINTER(ctypes.c_int64)),
        ("strides", ctypes.POINTER(ctypes.c_int64)),
        ("byte_offset", ctypes.c_uint64),
    ]


DLManagedTensorDeleter = ctypes.CFUNCTYPE(None, ctypes.c_void_p)


class DLManagedTensor(ctypes.Structure):
    #
    _fields_ = [
        ("dl_tensor", DLTensor),
        ("manager_ctx", ctypes.c_void_p),
        ("deleter", DLManagedTensorDeleter),
    ]


DLPACK_FLAG_BITMASK_READ_ONLY = 1 << 0
DLPACK_FLAG_BITMASK_IS_COPIED = 1 << 1

DLManagedTensorVersionedDeleter = ctypes.CFUNCTYPE(None, ctypes.c_void_p)


class DLManagedTensorVersioned(ctypes.Structure):
    #
    _fields_ = [
        ("version", DLPackVersion),
        ("manager_ctx", ctypes.c_void_p),
        ("deleter", DLManagedTensorDeleter),
        ("flags", ctypes.c_uint64),
        ("dl_tensor", DLTensor),
    ]


pyapi = ctypes.pythonapi

DLManagedTensor_p = ctypes.POINTER(DLManagedTensor)
DLManagedTensorVersioned_p = ctypes.POINTER(DLManagedTensorVersioned)

Py_IncRef = pyapi.Py_IncRef
Py_IncRef.restype = None
Py_IncRef.argtypes = [ctypes.py_object]

Py_DecRef = pyapi.Py_DecRef
Py_DecRef.restype = None
Py_DecRef.argtypes = [ctypes.py_object]

PyCapsule_Destructor = ctypes.PYFUNCTYPE(None, ctypes.c_void_p)
PyCapsule_Destructor.from_param = classmethod(
    lambda _cls, arg: (
        ctypes._CFuncPtr.from_param(arg) if arg is not None else None
    )
)

PyCapsule_New = pyapi.PyCapsule_New
PyCapsule_New.restype = ctypes.py_object
PyCapsule_New.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    PyCapsule_Destructor,
]

PyCapsule_IsValid = pyapi.PyCapsule_IsValid
PyCapsule_IsValid.restype = ctypes.c_int
PyCapsule_IsValid.argtypes = [ctypes.py_object, ctypes.c_char_p]

PyCapsule_SetName = pyapi.PyCapsule_SetName
PyCapsule_SetName.restype = ctypes.c_int
PyCapsule_SetName.argtypes = [ctypes.py_object, ctypes.c_char_p]

PyCapsule_GetPointer = pyapi.PyCapsule_GetPointer
PyCapsule_GetPointer.restype = ctypes.c_void_p
PyCapsule_GetPointer.argtypes = [ctypes.py_object, ctypes.c_char_p]

PyCapsule_SetContext = pyapi.PyCapsule_SetContext
PyCapsule_SetContext.restype = ctypes.c_int
PyCapsule_SetContext.argtypes = [ctypes.py_object, ctypes.c_void_p]

PyCapsule_GetContext = pyapi.PyCapsule_GetContext
PyCapsule_GetContext.restype = ctypes.c_void_p
PyCapsule_GetContext.argtypes = [ctypes.py_object]


def make_dl_version(major, minor):
    version = DLPackVersion()
    version.major = major
    version.minor = minor
    return version


def make_dl_datatype(typecode, itemsize):
    code = None
    bits = itemsize * 8
    lanes = 1
    if typecode == "?":
        code = DLDataTypeCode.kDLBool
    if typecode in "bhilqnp":
        code = DLDataTypeCode.kDLInt
    if typecode in "BHILQNP":
        code = DLDataTypeCode.kDLUInt
    if typecode in "efdg":
        code = DLDataTypeCode.kDLFloat
    if typecode in "FDG":
        code = DLDataTypeCode.kDLComplex
    if typecode == "G" and itemsize == 32:
        code = DLDataTypeCode.kDLFloat
        bits //= 2
        lanes *= 2
    datatype = DLDataType()
    datatype.code = code
    datatype.bits = bits
    datatype.lanes = lanes
    return datatype


def make_dl_shape(shape, order=None, strides=None):
    null = ctypes.cast(0, ctypes.POINTER(ctypes.c_int64))
    if isinstance(shape, int):
        shape = [shape]
    ndim = len(shape)
    if ndim == 0:
        shape = null
        strides = null
    else:
        shape = (ctypes.c_int64 * ndim)(*shape)
        if order == "C":
            size = 1
            strides = []
            for i in range(ndim - 1, -1, -1):
                strides.append(size)
                size *= shape[i]
            strides = (ctypes.c_int64 * ndim)(*strides)
        elif order == "F":
            size = 1
            strides = []
            for i in range(ndim):
                strides.append(size)
                size *= shape[i]
            strides = (ctypes.c_int64 * ndim)(*strides)
        elif strides is not None:
            strides = (ctypes.c_int64 * ndim)(*strides)
        else:
            strides = null
    return ndim, shape, strides


def make_dl_tensor(obj):
    try:
        data, size = obj.buffer_info()
        typecode = obj.typecode
        itemsize = obj.itemsize
    except AttributeError:
        data = obj.ctypes.data
        size = obj.size
        typecode = obj.dtype.char
        itemsize = obj.itemsize

    device = DLDevice(DLDeviceType.kDLCPU, 0)
    datatype = make_dl_datatype(typecode, itemsize)
    ndim, shape, strides = make_dl_shape(size)

    dltensor = DLTensor()
    dltensor.data = data if size > 0 else 0
    dltensor.device = device
    dltensor.ndim = ndim
    dltensor.dtype = datatype
    dltensor.shape = shape
    dltensor.strides = strides
    dltensor.byte_offset = 0
    return dltensor


def make_dl_manager_ctx(obj):
    py_obj = ctypes.py_object(obj)
    void_p = ctypes.c_void_p.from_buffer(py_obj)
    return void_p


@DLManagedTensorDeleter
def dl_managed_tensor_deleter(void_p):
    managed = ctypes.cast(void_p, DLManagedTensor_p)
    manager_ctx = managed.contents.manager_ctx
    py_obj = ctypes.cast(manager_ctx, ctypes.py_object)
    obj = py_obj.value
    del obj


@DLManagedTensorVersionedDeleter
def dl_managed_tensor_versioned_deleter(void_p):
    managed = ctypes.cast(void_p, DLManagedTensorVersioned_p)
    manager_ctx = managed.contents.manager_ctx
    py_obj = ctypes.cast(manager_ctx, ctypes.py_object)
    obj = py_obj.value
    del obj


def make_dl_managed_tensor(obj, version=0):
    if hasattr(sys, "getrefcount"):
        rc = sys.getrefcount(obj)
    if version >= 1:
        managed = DLManagedTensorVersioned()
        managed.version = make_dl_version(1, 0)
        managed.manager_ctx = make_dl_manager_ctx(obj)
        managed.deleter = dl_managed_tensor_versioned_deleter
        managed.flags = 0
        managed.dl_tensor = make_dl_tensor(obj)
    else:
        managed = DLManagedTensor()
        managed.dl_tensor = make_dl_tensor(obj)
        managed.manager_ctx = make_dl_manager_ctx(obj)
        managed.deleter = dl_managed_tensor_deleter
    if hasattr(sys, "getrefcount"):
        assert sys.getrefcount(obj) == rc + 1
    return managed


@PyCapsule_Destructor
def py_capsule_destructor(void_p):
    capsule = ctypes.cast(void_p, ctypes.py_object)
    used_py_capsule(capsule)
    nullptr = ctypes.c_void_p(None)
    context = PyCapsule_GetContext(capsule)
    if context != nullptr:
        PyCapsule_SetContext(capsule, nullptr)
        managed = ctypes.cast(context, ctypes.py_object)
        Py_DecRef(managed)


def make_py_capsule(managed, owned=True):
    if isinstance(managed, DLManagedTensorVersioned):
        py_capsule_name = b"dltensor_versioned"
    if isinstance(managed, DLManagedTensor):
        py_capsule_name = b"dltensor"
    destructor = None
    if owned:
        destructor = py_capsule_destructor
    pointer = ctypes.byref(managed)
    capsule = PyCapsule_New(pointer, py_capsule_name, destructor)
    if owned:
        context = ctypes.c_void_p.from_buffer(ctypes.py_object(managed))
        if sys.implementation.name == "cpython":
            assert context.value == id(managed)
        Py_IncRef(managed)
        PyCapsule_SetContext(capsule, context)
    return capsule


_used = {
    b"dltensor" + prefix: b"used_dltensor" + prefix
    for prefix in (b"_versioned", b"")
}


def used_py_capsule(capsule):
    for py_capsule_name, dl_managed_tensor_type_p in (
        (b"dltensor_versioned", DLManagedTensorVersioned_p),
        (b"dltensor", DLManagedTensor_p),
    ):
        if PyCapsule_IsValid(capsule, py_capsule_name):
            pointer = PyCapsule_GetPointer(capsule, py_capsule_name)
            managed = ctypes.cast(pointer, dl_managed_tensor_type_p)
            deleter = managed.contents.deleter
            if deleter:
                deleter(managed)
            PyCapsule_SetName(capsule, _used[py_capsule_name])
            break



from functools import reduce
from operator import mul

import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


def prod(sequence, start=1):
    return reduce(mul, sequence, start)


def skip_op(typecode, op):
    if typecode == "?":
        return True
    if typecode in "FDG":
        if op in (MPI.MAX, MPI.MIN):
            return True
    return False


def maxvalue(a):
    try:
        typecode = a.typecode
    except AttributeError:
        typecode = a.dtype.char
    if typecode == ("f"):
        return 1e30
    elif typecode == ("d"):
        return 1e300
    else:
        return 2 ** (a.itemsize * 7) - 1


@unittest.skipMPI("msmpi(<8.1.0)")
class BaseTestCCOBuf:
    #
    COMM = MPI.COMM_NULL

    def testBarrier(self):
        self.COMM.Ibarrier().Wait()

    def testBcast(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for root in range(size):
                check = arrayimpl.scalar(root)
                if rank == root:
                    buf = array(root, typecode, root)
                else:
                    buf = array(-1, typecode, root)
                self.COMM.Ibcast(buf.as_mpi(), root=root).Wait()
                for value in buf:
                    self.assertEqual(value, check)

    def testGather(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for root in range(size):
                check = arrayimpl.scalar(root)
                sbuf = array(root, typecode, root + 1)
                if rank == root:
                    rbuf = array(-1, typecode, (size, root + 1))
                else:
                    rbuf = array([], typecode)
                self.COMM.Igather(
                    sbuf.as_mpi(), rbuf.as_mpi(), root=root
                ).Wait()
                if rank == root:
                    for value in rbuf.flat:
                        self.assertEqual(value, check)

    def testScatter(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for root in range(size):
                check = arrayimpl.scalar(root)
                rbuf = array(-1, typecode, size)
                if rank == root:
                    sbuf = array(root, typecode, (size, size))
                else:
                    sbuf = array([], typecode)
                self.COMM.Iscatter(
                    sbuf.as_mpi(), rbuf.as_mpi(), root=root
                ).Wait()
                for value in rbuf:
                    self.assertEqual(value, check)

    def testAllgather(self):
        size = self.COMM.Get_size()
        self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for root in range(size):
                check = arrayimpl.scalar(root)
                sbuf = array(root, typecode, root + 1)
                rbuf = array(-1, typecode, (size, root + 1))
                self.COMM.Iallgather(sbuf.as_mpi(), rbuf.as_mpi()).Wait()
                for value in rbuf.flat:
                    self.assertEqual(value, check)

    def testAlltoall(self):
        size = self.COMM.Get_size()
        self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for root in range(size):
                check = arrayimpl.scalar(root)
                sbuf = array(root, typecode, (size, root + 1))
                rbuf = array(-1, typecode, (size, root + 1))
                self.COMM.Ialltoall(
                    sbuf.as_mpi(), rbuf.as_mpi_c(root + 1)
                ).Wait()
                for value in rbuf.flat:
                    self.assertEqual(value, check)

    def testReduce(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                for root in range(size):
                    sbuf = array(range(size), typecode)
                    rbuf = array(-1, typecode, size)
                    self.COMM.Ireduce(
                        sbuf.as_mpi(), rbuf.as_mpi(), op, root
                    ).Wait()
                    max_val = maxvalue(rbuf)
                    for i, value in enumerate(rbuf):
                        if rank != root:
                            check = arrayimpl.scalar(-1)
                            self.assertEqual(value, check)
                            continue
                        if op == MPI.SUM:
                            if (i * size) < max_val:
                                self.assertAlmostEqual(value, i * size)
                        elif op == MPI.PROD:
                            if (i**size) < max_val:
                                self.assertAlmostEqual(value, i**size)
                        elif op == MPI.MAX:
                            self.assertEqual(value, i)
                        elif op == MPI.MIN:
                            self.assertEqual(value, i)

    def testAllreduce(self):
        size = self.COMM.Get_size()
        self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
                if skip_op(typecode, op):
                    continue
                sbuf = array(range(size), typecode)
                rbuf = array(0, typecode, size)
                self.COMM.Iallreduce(sbuf.as_mpi(), rbuf.as_mpi(), op).Wait()
                max_val = maxvalue(rbuf)
                for i, value in enumerate(rbuf):
                    if op == MPI.SUM:
                        if (i * size) < max_val:
                            self.assertAlmostEqual(value, i * size)
                    elif op == MPI.PROD:
                        if (i**size) < max_val:
                            self.assertAlmostEqual(value, i**size)
                    elif op == MPI.MAX:
                        self.assertEqual(value, i)
                    elif op == MPI.MIN:
                        self.assertEqual(value, i)

    @unittest.skipMPI("openmpi(<=1.8.3)")
    def testReduceScatter(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
                if skip_op(typecode, op):
                    continue
                rcnt = list(range(1, size + 1))
                sbuf = array([rank + 1] * sum(rcnt), typecode)
                rbuf = array(-1, typecode, rank + 1)
                self.COMM.Ireduce_scatter(
                    sbuf.as_mpi(), rbuf.as_mpi(), None, op
                ).Wait()
                max_val = maxvalue(rbuf)
                for value in rbuf:
                    if op == MPI.SUM:
                        redval = sum(range(size)) + size
                        if redval < max_val:
                            self.assertAlmostEqual(value, redval)
                    elif op == MPI.PROD:
                        redval = prod(range(1, size + 1))
                        if redval < max_val:
                            self.assertAlmostEqual(value, redval)
                    elif op == MPI.MAX:
                        self.assertEqual(value, size)
                    elif op == MPI.MIN:
                        self.assertEqual(value, 1)
                rbuf = array(-1, typecode, rank + 1)
                self.COMM.Ireduce_scatter(
                    sbuf.as_mpi(), rbuf.as_mpi(), rcnt, op
                ).Wait()
                max_val = maxvalue(rbuf)
                for value in rbuf:
                    if op == MPI.SUM:
                        redval = sum(range(size)) + size
                        if redval < max_val:
                            self.assertAlmostEqual(value, redval)
                    elif op == MPI.PROD:
                        redval = prod(range(1, size + 1))
                        if redval < max_val:
                            self.assertAlmostEqual(value, redval)
                    elif op == MPI.MAX:
                        self.assertEqual(value, size)
                    elif op == MPI.MIN:
                        self.assertEqual(value, 1)

    def testReduceScatterBlock(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
                if skip_op(typecode, op):
                    continue
                for rcnt in range(1, size + 1):
                    sbuf = array([rank] * rcnt * size, typecode)
                    rbuf = array(-1, typecode, rcnt)
                    if op == MPI.PROD:
                        sbuf = array([rank + 1] * rcnt * size, typecode)
                    self.COMM.Ireduce_scatter_block(
                        sbuf.as_mpi(), rbuf.as_mpi(), op
                    ).Wait()
                    max_val = maxvalue(rbuf)
                    v_sum = (size * (size - 1)) / 2
                    v_prod = 1
                    for i in range(1, size + 1):
                        v_prod *= i
                    v_max = size - 1
                    v_min = 0
                    for value in rbuf:
                        if op == MPI.SUM:
                            if v_sum <= max_val:
                                self.assertAlmostEqual(value, v_sum)
                        elif op == MPI.PROD:
                            if v_prod <= max_val:
                                self.assertAlmostEqual(value, v_prod)
                        elif op == MPI.MAX:
                            self.assertEqual(value, v_max)
                        elif op == MPI.MIN:
                            self.assertEqual(value, v_min)

    def testScan(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        # --
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                sbuf = array(range(size), typecode)
                rbuf = array(0, typecode, size)
                self.COMM.Iscan(sbuf.as_mpi(), rbuf.as_mpi(), op).Wait()
                max_val = maxvalue(rbuf)
                for i, value in enumerate(rbuf):
                    if op == MPI.SUM:
                        if (i * (rank + 1)) < max_val:
                            self.assertAlmostEqual(value, i * (rank + 1))
                    elif op == MPI.PROD:
                        if (i ** (rank + 1)) < max_val:
                            self.assertAlmostEqual(value, i ** (rank + 1))
                    elif op == MPI.MAX:
                        self.assertEqual(value, i)
                    elif op == MPI.MIN:
                        self.assertEqual(value, i)

    @unittest.skipMPI("openmpi(<=1.8.1)")
    def testExscan(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                sbuf = array(range(size), typecode)
                rbuf = array(0, typecode, size)
                self.COMM.Iexscan(sbuf.as_mpi(), rbuf.as_mpi(), op).Wait()
                if rank == 1:
                    for i, value in enumerate(rbuf):
                        self.assertEqual(value, i)
                elif rank > 1:
                    max_val = maxvalue(rbuf)
                    for i, value in enumerate(rbuf):
                        if op == MPI.SUM:
                            if (i * rank) < max_val:
                                self.assertAlmostEqual(value, i * rank)
                        elif op == MPI.PROD:
                            if (i**rank) < max_val:
                                self.assertAlmostEqual(value, i**rank)
                        elif op == MPI.MAX:
                            self.assertEqual(value, i)
                        elif op == MPI.MIN:
                            self.assertEqual(value, i)

    def testBcastTypeIndexed(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            datatype = array.TypeMap[typecode]
            for root in range(size):
                #
                if rank == root:
                    buf = array(range(10), typecode).as_raw()
                else:
                    buf = array(-1, typecode, 10).as_raw()
                indices = list(range(0, len(buf), 2))
                newtype = datatype.Create_indexed_block(1, indices)
                newtype.Commit()
                newbuf = (buf, 1, newtype)
                self.COMM.Ibcast(newbuf, root=root).Wait()
                newtype.Free()
                if rank != root:
                    for i, value in enumerate(buf):
                        check = arrayimpl.scalar(-1 if (i % 2) else i)
                        self.assertEqual(value, check)

                #
                if rank == root:
                    buf = array(range(10), typecode).as_raw()
                else:
                    buf = array(-1, typecode, 10).as_raw()
                indices = list(range(1, len(buf), 2))
                newtype = datatype.Create_indexed_block(1, indices)
                newtype.Commit()
                newbuf = (buf, 1, newtype)
                self.COMM.Ibcast(newbuf, root).Wait()
                newtype.Free()
                if rank != root:
                    for i, value in enumerate(buf):
                        check = arrayimpl.scalar(-1 if not (i % 2) else i)
                        self.assertEqual(value, check)


@unittest.skipMPI("msmpi(<8.1.0)")
class BaseTestCCOBufInplace:
    #
    def testGather(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for root in range(size):
                check = arrayimpl.scalar(root)
                count = root + 3
                if rank == root:
                    sbuf = MPI.IN_PLACE
                    buf = array(-1, typecode, (size, count))
                    # buf.flat[(rank*count):((rank+1)*count)] = \
                    #    array(root, typecode, count)
                    s, e = rank * count, (rank + 1) * count
                    for i in range(s, e):
                        buf.flat[i] = check
                    rbuf = buf.as_mpi()
                else:
                    buf = array(root, typecode, count)
                    sbuf = buf.as_mpi()
                    rbuf = None
                self.COMM.Igather(sbuf, rbuf, root=root).Wait()
                for value in buf.flat:
                    self.assertEqual(value, check)
                if rank == root:
                    sbuf = None
                self.COMM.Igather(sbuf, rbuf, root=root).Wait()
                for value in buf.flat:
                    self.assertEqual(value, check)

    @unittest.skipMPI("msmpi(==10.0.0)")
    def testScatter(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for root in range(size):
                check = arrayimpl.scalar(root)
                for count in range(1, 10):
                    if rank == root:
                        buf = array(root, typecode, (size, count))
                        sbuf = buf.as_mpi()
                        rbuf = MPI.IN_PLACE
                    else:
                        buf = array(-1, typecode, count)
                        sbuf = None
                        rbuf = buf.as_mpi()
                    self.COMM.Iscatter(sbuf, rbuf, root=root).Wait()
                    for value in buf.flat:
                        self.assertEqual(value, check)
                    if rank == root:
                        rbuf = None
                    self.COMM.Iscatter(sbuf, rbuf, root=root).Wait()
                    for value in buf.flat:
                        self.assertEqual(value, check)

    def testAllgather(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            for count in range(1, 10):
                check = arrayimpl.scalar(count)
                buf = array(-1, typecode, (size, count))
                s, e = rank * count, (rank + 1) * count
                for i in range(s, e):
                    buf.flat[i] = check
                self.COMM.Iallgather(MPI.IN_PLACE, buf.as_mpi()).Wait()
                for value in buf.flat:
                    self.assertEqual(value, check)
                self.COMM.Iallgather(None, buf.as_mpi()).Wait()
                for value in buf.flat:
                    self.assertEqual(value, check)

    def testReduce(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                for root in range(size):
                    if rank == root:
                        buf = array(range(size), typecode)
                        sbuf = MPI.IN_PLACE
                        rbuf = buf.as_mpi()
                    else:
                        buf = array(range(size), typecode)
                        buf2 = array(range(size), typecode)
                        sbuf = buf.as_mpi()
                        rbuf = buf2.as_mpi()
                    self.COMM.Ireduce(sbuf, rbuf, op, root).Wait()
                    if rank == root:
                        max_val = maxvalue(buf)
                        for i, value in enumerate(buf):
                            if op == MPI.SUM:
                                if (i * size) < max_val:
                                    self.assertAlmostEqual(value, i * size)
                            elif op == MPI.PROD:
                                if (i**size) < max_val:
                                    self.assertAlmostEqual(value, i**size)
                            elif op == MPI.MAX:
                                self.assertEqual(value, i)
                            elif op == MPI.MIN:
                                self.assertEqual(value, i)

    def testAllreduce(self):
        size = self.COMM.Get_size()
        self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
                if skip_op(typecode, op):
                    continue
                buf = array(range(size), typecode)
                sbuf = MPI.IN_PLACE
                rbuf = buf.as_mpi()
                self.COMM.Iallreduce(sbuf, rbuf, op).Wait()
                max_val = maxvalue(buf)
                for i, value in enumerate(buf):
                    if op == MPI.SUM:
                        if (i * size) < max_val:
                            self.assertAlmostEqual(value, i * size)
                    elif op == MPI.PROD:
                        if (i**size) < max_val:
                            self.assertAlmostEqual(value, i**size)
                    elif op == MPI.MAX:
                        self.assertEqual(value, i)
                    elif op == MPI.MIN:
                        self.assertEqual(value, i)

    @unittest.skipMPI("openmpi(<=1.8.6)")
    def testReduceScatterBlock(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
                if skip_op(typecode, op):
                    continue
                for rcnt in range(size):
                    if op == MPI.PROD:
                        rbuf = array([rank + 1] * rcnt * size, typecode)
                    else:
                        rbuf = array([rank] * rcnt * size, typecode)
                    self.COMM.Ireduce_scatter_block(
                        MPI.IN_PLACE, rbuf.as_mpi(), op
                    ).Wait()
                    max_val = maxvalue(rbuf)
                    for i, value in enumerate(rbuf):
                        if i >= rcnt:
                            if op == MPI.PROD:
                                self.assertEqual(value, rank + 1)
                            else:
                                self.assertEqual(value, rank)
                        else:
                            if op == MPI.SUM:
                                redval = sum(range(size))
                                if redval < max_val:
                                    self.assertAlmostEqual(value, redval)
                            elif op == MPI.PROD:
                                redval = prod(range(1, size + 1))
                                if redval < max_val:
                                    self.assertAlmostEqual(value, redval)
                            elif op == MPI.MAX:
                                self.assertEqual(value, size - 1)
                            elif op == MPI.MIN:
                                self.assertEqual(value, 0)

    @unittest.skipMPI("openmpi(<=1.8.6)")
    def testReduceScatter(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
                if skip_op(typecode, op):
                    continue
                rcnt = list(range(1, size + 1))
                if op == MPI.PROD:
                    rbuf = array([rank + 1] * sum(rcnt), typecode)
                else:
                    rbuf = array([rank] * sum(rcnt), typecode)
                self.COMM.Ireduce_scatter(
                    MPI.IN_PLACE, rbuf.as_mpi(), rcnt, op
                ).Wait()
                max_val = maxvalue(rbuf)
                for i, value in enumerate(rbuf):
                    if i >= rcnt[rank]:
                        if op == MPI.PROD:
                            self.assertEqual(value, rank + 1)
                        else:
                            self.assertEqual(value, rank)
                    else:
                        if op == MPI.SUM:
                            redval = sum(range(size))
                            if redval < max_val:
                                self.assertAlmostEqual(value, redval)
                        elif op == MPI.PROD:
                            redval = prod(range(1, size + 1))
                            if redval < max_val:
                                self.assertAlmostEqual(value, redval)
                        elif op == MPI.MAX:
                            self.assertEqual(value, size - 1)
                        elif op == MPI.MIN:
                            self.assertEqual(value, 0)

    @unittest.skipMPI("openmpi(<=1.8.4)")
    def testScan(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        # --
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                buf = array(range(size), typecode)
                self.COMM.Iscan(MPI.IN_PLACE, buf.as_mpi(), op).Wait()
                max_val = maxvalue(buf)
                for i, value in enumerate(buf):
                    if op == MPI.SUM:
                        if (i * (rank + 1)) < max_val:
                            self.assertAlmostEqual(value, i * (rank + 1))
                    elif op == MPI.PROD:
                        if (i ** (rank + 1)) < max_val:
                            self.assertAlmostEqual(value, i ** (rank + 1))
                    elif op == MPI.MAX:
                        self.assertEqual(value, i)
                    elif op == MPI.MIN:
                        self.assertEqual(value, i)

    @unittest.skipMPI("openmpi(<=1.8.4)")
    def testExscan(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            # segfault as of OpenMPI 4.1.1
            if unittest.is_mpi_gpu("openmpi", array):
                continue
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                buf = array(range(size), typecode)
                try:
                    self.COMM.Iexscan(MPI.IN_PLACE, buf.as_mpi(), op).Wait()
                except NotImplementedError:
                    self.skipTest("mpi-exscan")
                if rank == 1:
                    for i, value in enumerate(buf):
                        self.assertEqual(value, i)
                elif rank > 1:
                    max_val = maxvalue(buf)
                    for i, value in enumerate(buf):
                        if op == MPI.SUM:
                            if (i * rank) < max_val:
                                self.assertAlmostEqual(value, i * rank)
                        elif op == MPI.PROD:
                            if (i**rank) < max_val:
                                self.assertAlmostEqual(value, i**rank)
                        elif op == MPI.MAX:
                            self.assertEqual(value, i)
                        elif op == MPI.MIN:
                            self.assertEqual(value, i)


class TestCCOBufSelf(BaseTestCCOBuf, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestCCOBufWorld(BaseTestCCOBuf, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestCCOBufInplaceSelf(BaseTestCCOBufInplace, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestCCOBufInplaceWorld(BaseTestCCOBufInplace, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD

    @unittest.skipMPI("mvapich(<3.0.0)", MPI.COMM_WORLD.Get_size() > 1)
    def testReduceScatter(self):
        super().testReduceScatter()


class TestCCOBufSelfDup(TestCCOBufSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


class TestCCOBufWorldDup(TestCCOBufWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


try:
    MPI.COMM_SELF.Ibarrier().Wait()
except NotImplementedError:
    unittest.disable(BaseTestCCOBuf, "mpi-nbc")
    unittest.disable(BaseTestCCOBufInplace, "mpi-nbc")


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI


class TestTopoConstructor(unittest.TestCase):
    #
    def testConstructorTopocomm(self):
        comm = MPI.COMM_SELF
        with self.assertRaises(TypeError):
            MPI.Topocomm(comm)

    def testConstructorCartcomm(self):
        comm = MPI.COMM_SELF
        cart = comm.Create_cart([1])
        with self.assertRaises(TypeError):
            MPI.Graphcomm(cart)
        with self.assertRaises(TypeError):
            MPI.Distgraphcomm(cart)
        cart.Free()

    def testConstructorGraphcomm(self):
        comm = MPI.COMM_SELF
        graph = comm.Create_graph([0, 1], [0])
        with self.assertRaises(TypeError):
            MPI.Cartcomm(graph)
        with self.assertRaises(TypeError):
            MPI.Distgraphcomm(graph)
        graph.Free()

    def testConstructorDistGraphcomm(self):
        comm = MPI.COMM_SELF
        try:
            distgraph = comm.Create_dist_graph([], [], [])
        except NotImplementedError:
            self.skipTest("mpi-comm-create_dist_graph")
        with self.assertRaises(TypeError):
            MPI.Cartcomm(distgraph)
        with self.assertRaises(TypeError):
            MPI.Graphcomm(distgraph)
        distgraph.Free()
        with self.assertRaises(ValueError):
            comm.Create_dist_graph_adjacent(
                [0],
                [0],
                MPI.WEIGHTS_EMPTY,
                MPI.WEIGHTS_EMPTY,
            )


class BaseTestTopo:
    #
    COMM = MPI.COMM_NULL

    def checkHandle(self, oldcomm):
        cint = oldcomm.toint()
        if cint != -1:
            newcomm = MPI.Comm.fromint(cint)
            self.assertEqual(newcomm, oldcomm)
            self.assertEqual(type(newcomm), type(oldcomm))
        fint = oldcomm.py2f()
        if fint != -1:
            newcomm = MPI.Comm.f2py(fint)
            self.assertEqual(newcomm, oldcomm)
            self.assertEqual(type(newcomm), type(oldcomm))

    def testCartcomm(self):
        comm = self.COMM
        size = comm.Get_size()
        comm.Get_rank()
        for ndim in (1, 2, 3, 4, 5):
            dims = MPI.Compute_dims(size, [0] * ndim)
            periods = [True] * len(dims)
            topo = comm.Create_cart(dims, periods=periods)
            self.assertTrue(topo.is_topo)
            self.assertEqual(topo.topology, MPI.CART)
            self.checkHandle(topo)
            self.assertEqual(topo.dim, len(dims))
            self.assertEqual(topo.ndim, len(dims))
            coords = topo.coords
            self.assertEqual(coords, topo.Get_coords(topo.rank))
            self.assertEqual(topo.topo, (dims, periods, coords))
            neighbors = []
            coordinates = topo.Get_coords(topo.rank)
            for i in range(ndim):
                for d in (-1, +1):
                    coord = list(coordinates)
                    coord[i] = (coord[i] + d) % dims[i]
                    neigh = topo.Get_cart_rank(coord)
                    self.assertEqual(coord, topo.Get_coords(neigh))
                    _source, dest = topo.Shift(i, d)
                    self.assertEqual(neigh, dest)
                    neighbors.append(neigh)
            self.assertEqual(topo.indegree, len(neighbors))
            self.assertEqual(topo.outdegree, len(neighbors))
            self.assertEqual(topo.inedges, neighbors)
            self.assertEqual(topo.outedges, neighbors)
            inedges, outedges = topo.inoutedges
            self.assertEqual(inedges, neighbors)
            self.assertEqual(outedges, neighbors)
            if ndim == 1:
                topo.Free()
                continue
            for i in range(ndim):
                rem_dims = [1] * ndim
                rem_dims[i] = 0
                sub = topo.Sub(rem_dims)
                if sub != MPI.COMM_NULL:
                    self.assertEqual(sub.dim, ndim - 1)
                    dims = topo.dims
                    del dims[i]
                    self.assertEqual(sub.dims, dims)
                    sub.Free()
            topo.Free()
        with self.assertRaises(ValueError):
            topo = comm.Create_cart([comm.Get_size()], [])
        with self.assertRaises(ValueError):
            topo = comm.Create_cart([comm.Get_size()], [0, 0])

    @unittest.skipMPI("MPI(<2.0)")
    def testCartcommZeroDim(self):
        comm = self.COMM
        topo = comm.Create_cart([])
        if topo == MPI.COMM_NULL:
            return
        self.assertEqual(topo.dim, 0)
        self.assertEqual(topo.dims, [])
        self.assertEqual(topo.periods, [])
        self.assertEqual(topo.coords, [])
        rank = topo.Get_cart_rank([])
        self.assertEqual(rank, 0)
        inedges, outedges = topo.inoutedges
        self.assertEqual(inedges, [])
        self.assertEqual(outedges, [])
        topo.Free()

    def testGraphcomm(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        index, edges = [0], []
        for i in range(size):
            pos = index[-1]
            index.append(pos + 2)
            edges.extend(((i - 1) % size, (i + 1) % size))
        topo = comm.Create_graph(index[1:], edges)
        self.assertTrue(topo.is_topo)
        self.assertEqual(topo.topology, MPI.GRAPH)
        self.checkHandle(topo)
        topo.Free()
        topo = comm.Create_graph(index, edges)
        self.assertEqual(topo.dims, (len(index) - 1, len(edges)))
        self.assertEqual(topo.nnodes, len(index) - 1)
        self.assertEqual(topo.nedges, len(edges))
        self.assertEqual(topo.index, index[1:])
        self.assertEqual(topo.edges, edges)
        self.assertEqual(topo.topo, (index[1:], edges))
        neighbors = edges[index[rank] : index[rank + 1]]
        self.assertEqual(neighbors, topo.neighbors)
        self.assertEqual(len(neighbors), topo.nneighbors)
        for rank in range(size):
            neighs = topo.Get_neighbors(rank)
            self.assertEqual(neighs, [(rank - 1) % size, (rank + 1) % size])
        self.assertEqual(topo.indegree, len(neighbors))
        self.assertEqual(topo.outdegree, len(neighbors))
        self.assertEqual(topo.inedges, neighbors)
        self.assertEqual(topo.outedges, neighbors)
        inedges, outedges = topo.inoutedges
        self.assertEqual(inedges, neighbors)
        self.assertEqual(outedges, neighbors)
        topo.Free()

    @unittest.skipMPI("msmpi")
    def testDistgraphcommAdjacent(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        try:
            topo = comm.Create_dist_graph_adjacent(None, None)
            topo.Free()
        except NotImplementedError:
            self.skipTest("mpi-comm-create_dist_graph_adjacent")
        #
        sources = [(rank - 2) % size, (rank - 1) % size]
        destinations = [(rank + 1) % size, (rank + 2) % size]
        topo = comm.Create_dist_graph_adjacent(sources, destinations)
        self.assertTrue(topo.is_topo)
        self.assertEqual(topo.topology, MPI.DIST_GRAPH)
        self.checkHandle(topo)
        self.assertEqual(topo.Get_dist_neighbors_count(), (2, 2, False))
        self.assertEqual(
            topo.Get_dist_neighbors(), (sources, destinations, None)
        )
        self.assertEqual(topo.indegree, len(sources))
        self.assertEqual(topo.outdegree, len(destinations))
        self.assertEqual(topo.inedges, sources)
        self.assertEqual(topo.outedges, destinations)
        inedges, outedges = topo.inoutedges
        self.assertEqual(inedges, sources)
        self.assertEqual(outedges, destinations)
        topo.Free()
        #
        sourceweights = [1, 2]
        destweights = [3, 4]
        weights = (sourceweights, destweights)
        topo = comm.Create_dist_graph_adjacent(
            sources, destinations, sourceweights, destweights
        )
        self.assertEqual(topo.Get_dist_neighbors_count(), (2, 2, True))
        self.assertEqual(
            topo.Get_dist_neighbors(), (sources, destinations, weights)
        )
        topo.Free()
        #
        topo = comm.Create_dist_graph_adjacent(
            sources, None, MPI.UNWEIGHTED, None
        )
        self.assertEqual(topo.Get_dist_neighbors_count(), (2, 0, False))
        self.assertEqual(topo.Get_dist_neighbors(), (sources, [], None))
        topo.Free()
        topo = comm.Create_dist_graph_adjacent(
            None, destinations, None, MPI.UNWEIGHTED
        )
        self.assertEqual(topo.Get_dist_neighbors_count(), (0, 2, False))
        self.assertEqual(topo.Get_dist_neighbors(), ([], destinations, None))
        topo.Free()
        if MPI.Get_version() < (3, 0):
            return
        topo = comm.Create_dist_graph_adjacent(
            [], [], MPI.WEIGHTS_EMPTY, MPI.WEIGHTS_EMPTY
        )
        self.assertEqual(topo.Get_dist_neighbors_count(), (0, 0, True))
        self.assertEqual(topo.Get_dist_neighbors(), ([], [], ([], [])))
        topo.Free()

    @unittest.skipMPI("msmpi")
    def testDistgraphcomm(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        #
        try:
            topo = comm.Create_dist_graph([], [], [], MPI.UNWEIGHTED)
            topo.Free()
        except NotImplementedError:
            self.skipTest("mpi-comm-create_dist_graph")
        #
        sources = [rank]
        degrees = [3]
        destinations = [(rank - 1) % size, rank, (rank + 1) % size]
        topo = comm.Create_dist_graph(
            sources, degrees, destinations, MPI.UNWEIGHTED
        )
        self.assertTrue(topo.is_topo)
        self.assertEqual(topo.topology, MPI.DIST_GRAPH)
        self.checkHandle(topo)
        self.assertEqual(topo.Get_dist_neighbors_count(), (3, 3, False))
        topo.Free()
        weights = list(range(1, 4))
        topo = comm.Create_dist_graph(sources, degrees, destinations, weights)
        self.assertEqual(topo.Get_dist_neighbors_count(), (3, 3, True))
        topo.Free()

    def testCartMap(self):
        comm = self.COMM
        size = comm.Get_size()
        for ndim in (1, 2, 3, 4, 5):
            for periods in (None, True, False):
                dims = MPI.Compute_dims(size, ndim)
                topo = comm.Create_cart(dims, periods, reorder=True)
                rank = comm.Cart_map(dims, periods)
                self.assertEqual(topo.Get_rank(), rank)
                topo.Free()

    def testGraphMap(self):
        comm = self.COMM
        size = comm.Get_size()
        index, edges = [0], []
        for i in range(size):
            pos = index[-1]
            index.append(pos + 2)
            edges.extend(((i - 1) % size, (i + 1) % size))
        # Version 1
        topo = comm.Create_graph(index, edges, reorder=True)
        rank = comm.Graph_map(index, edges)
        self.assertEqual(topo.Get_rank(), rank)
        topo.Free()
        # Version 2
        topo = comm.Create_graph(index[1:], edges, reorder=True)
        rank = comm.Graph_map(index[1:], edges)
        self.assertEqual(topo.Get_rank(), rank)
        topo.Free()


class TestTopoSelf(BaseTestTopo, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestTopoWorld(BaseTestTopo, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestTopoSelfDup(TestTopoSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


class TestTopoWorldDup(TestTopoWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


if __name__ == "__main__":
    unittest.main()



# mpiexec -n 4 python ex-2.16.py

# Jacobi code
# version of parallel code using sendrecv and null processes.

# --------------------------------------------------------------------

from mpi4py import MPI

try:
    import numpy
except ImportError:
    raise SystemExit from None

# --------------------------------------------------------------------

n = 5 * MPI.COMM_WORLD.Get_size()

# compute number of processes and myrank
p = MPI.COMM_WORLD.Get_size()
myrank = MPI.COMM_WORLD.Get_rank()

# compute size of local block
m = n // p
if myrank < (n - p * m):
    m = m + 1

# compute neighbors
if myrank == 0:
    left = MPI.PROC_NULL
else:
    left = myrank - 1
if myrank == p - 1:
    right = MPI.PROC_NULL
else:
    right = myrank + 1

# allocate local arrays
A = numpy.empty((n + 2, m + 2), dtype="d", order="f")
B = numpy.empty((n, m), dtype="d", order="f")

A.fill(1)
A[0, :] = A[-1, :] = 0
A[:, 0] = A[:, -1] = 0

# main loop
converged = False
while not converged:
    # compute,  B = 0.25 * ( N + S + E + W)
    N, S = A[:-2, 1:-1], A[2:, 1:-1]
    E, W = A[1:-1, :-2], A[1:-1, 2:]
    numpy.add(N, S, B)
    numpy.add(E, B, B)
    numpy.add(W, B, B)
    B *= 0.25
    A[1:-1, 1:-1] = B
    # communicate
    tag = 0
    MPI.COMM_WORLD.Sendrecv(
        [B[:, -1], MPI.DOUBLE], right, tag, [A[:, 0], MPI.DOUBLE], left, tag
    )
    MPI.COMM_WORLD.Sendrecv(
        (B[:, 0], MPI.DOUBLE), left, tag, (A[:, -1], MPI.DOUBLE), right, tag
    )
    # convergence
    myconv = numpy.allclose(B, 0)
    loc_conv = numpy.asarray(myconv, dtype="i")
    glb_conv = numpy.asarray(0, dtype="i")
    MPI.COMM_WORLD.Allreduce(
        [loc_conv, MPI.INT], [glb_conv, MPI.INT], op=MPI.LAND
    )
    converged = bool(glb_conv)

# --------------------------------------------------------------------



import os
import pathlib
import shlex
import shutil
import signal
import string
import subprocess
import sys
import tempfile
import unittest
import warnings
import zipfile

import mpi4py

on_ci = any((
    os.environ.get("GITHUB_ACTIONS") == "true",
    os.environ.get("TF_BUILD") == "True",
    os.environ.get("CIRCLECI") == "true",
))


def find_executable(exe):
    command = shlex.split(exe)
    executable = shutil.which(command[0])
    if executable:
        command[0] = executable
        return shlex.join(command)
    return None


def find_mpiexec(mpiexec="mpiexec"):
    mpiexec = os.environ.get("MPIEXEC") or mpiexec
    mpiexec = find_executable(mpiexec)
    if not mpiexec and sys.platform.startswith("win"):
        I_MPI_DIR = pathlib.Path(os.environ.get("I_MPI_DIR", ""))
        mpiexec = str(I_MPI_DIR / "bin" / "mpiexec.exe")
        mpiexec = shutil.which(mpiexec)
        if mpiexec:
            mpiexec = shlex.quote(mpiexec)
    if not mpiexec and sys.platform.startswith("win"):
        MSMPI_BIN = pathlib.Path(os.environ.get("MSMPI_BIN", ""))
        mpiexec = str(MSMPI_BIN / "mpiexec.exe")
        mpiexec = shutil.which(mpiexec)
        if mpiexec:
            mpiexec = shlex.quote(mpiexec)
    return mpiexec


def launcher(np):
    mpiexec = find_mpiexec()
    python = shlex.quote(sys.executable)
    if "coverage" in sys.modules:
        python += " -m coverage run -p"
    module = "mpi4py.run -rc threads=False"
    command = f"{mpiexec} -n {np} {python} -m {module}"
    return shlex.split(command)


def execute(np, cmd, args=""):
    mpi4pyroot = pathlib.Path(mpi4py.__path__[0]).resolve().parent
    pythonpath = os.environ.get("PYTHONPATH", "").split(os.pathsep)
    pythonpath.insert(0, str(mpi4pyroot))

    mpiexec = launcher(np)
    if isinstance(cmd, str):
        cmd = shlex.split(cmd)
    if isinstance(args, str):
        args = shlex.split(args)
    command = mpiexec + cmd + args

    env = os.environ.copy()
    env["PYTHONPATH"] = os.pathsep.join(pythonpath)
    env["PYTHONUNBUFFERED"] = "1"

    p = subprocess.Popen(
        command,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate()
    return p.returncode, stdout.decode(), stderr.decode()


@unittest.skipIf(not find_mpiexec(), "mpiexec")
class BaseTestRun(unittest.TestCase):
    #

    def assertFailure(self, status, expected):
        if on_ci and status in {221, 222}:
            with warnings.catch_warnings():
                warnings.simplefilter("always")
                warnings.warn(
                    f"expecting status {expected}, got {status}",
                    RuntimeWarning,
                    2,
                )
            return
        if isinstance(expected, int):
            self.assertEqual(status, expected)
        else:
            self.assertIn(status, expected)

    def assertMPIAbort(self, stdout, stderr, message=None):
        patterns = (
            "MPI_Abort",  # MPICH
            "MPI_ABORT",  # Open MPI
            "aborting MPI_COMM_WORLD",  # Microsoft MPI
        )
        aborted = any(
            mpiabort in output
            for output in (stdout, stderr)
            for mpiabort in patterns
        )
        if aborted:
            if message is not None and not on_ci:
                self.assertIn(message, stderr)
            return
        if not (stdout or stderr) or on_ci:
            with warnings.catch_warnings():
                warnings.simplefilter("always")
                warnings.warn(
                    "expecting MPI_Abort() message in stdout/stderr",
                    RuntimeWarning,
                    2,
                )
            return
        raise self.failureException(
            "expecting MPI_Abort() message in stdout/stderr:\n"
            f"[stdout]:\n{stdout}\n[stderr]:\n{stderr}\n"
        )


class TestRunScript(BaseTestRun):
    pyfile = "run-script.py"

    def execute(self, args="", np=1):
        dirname = pathlib.Path(__file__).resolve().parent
        script = dirname / self.pyfile
        return execute(np, shlex.quote(str(script)), args)

    def testSuccess(self):
        success = "Hello, World!"
        for np in (1, 2):
            status, stdout, stderr = self.execute(np=np)
            self.assertEqual(status, 0)
            self.assertEqual(stdout.count(success), np)
            self.assertEqual(stderr, "")

    def testException(self):
        message = string.ascii_uppercase
        excmess = f"RuntimeError: {message}"
        for np in (1, 2):
            for rank in range(np):
                args = ["--rank", str(rank), "--exception", message]
                status, stdout, stderr = self.execute(args, np)
                self.assertFailure(status, 1)
                self.assertMPIAbort(stdout, stderr, excmess)

    def testSysExitCode(self):
        errcode = 7
        for np in (1, 2):
            for r in sorted({0, np - 1}):
                args = ["--rank", str(r), "--sys-exit", str(errcode)]
                status, stdout, stderr = self.execute(args, np)
                self.assertFailure(status, (errcode, 1))
                self.assertMPIAbort(stdout, stderr)
                self.assertNotIn("Traceback", stderr)

    def testSysExitMess(self):
        exitmsg = string.ascii_uppercase
        for np in (1, 2):
            for r in sorted({0, np - 1}):
                args = ["--rank", str(r), "--sys-exit-msg", exitmsg]
                status, stdout, stderr = self.execute(args, np)
                self.assertFailure(status, 1)
                self.assertMPIAbort(stdout, stderr, exitmsg)
                self.assertNotIn("Traceback", stderr)

    def testInterrupt(self):
        excmess = "KeyboardInterrupt"
        for np in (1, 2):
            for rank in range(np):
                args = ["--rank", str(rank), "--interrupt"]
                status, stdout, stderr = self.execute(args, np)
                self.assertFailure(status, signal.SIGINT + 128)
                self.assertMPIAbort(stdout, stderr, excmess)


class TestRunDirectory(TestRunScript):
    directory = "run-directory"

    @classmethod
    def setUpClass(cls):
        cls.tempdir = pathlib.Path(tempfile.mkdtemp())
        cls.directory = cls.tempdir / cls.directory
        cls.directory.mkdir(parents=True)
        topdir = pathlib.Path(__file__).parent
        script = topdir / super().pyfile
        pymain = cls.directory / "__main__.py"
        shutil.copy(script, pymain)
        cls.pyfile = cls.directory

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)


class TestRunZipFile(TestRunScript):
    zipfile = "run-zipfile.zip"

    @classmethod
    def setUpClass(cls):
        cls.tempdir = pathlib.Path(tempfile.mkdtemp())
        cls.zipfile = cls.tempdir / cls.zipfile
        topdir = pathlib.Path(__file__).parent
        script = topdir / super().pyfile
        with zipfile.ZipFile(cls.zipfile, "w") as f:
            f.write(script, "__main__.py")
        cls.pyfile = cls.zipfile

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)


class TestRunModule(BaseTestRun):
    def execute(self, module, np=1):
        return execute(np, "-m", module)

    def testSuccess(self):
        module = "mpi4py.bench --no-threads helloworld"
        message = "Hello, World!"
        for np in (1, 2):
            status, stdout, stderr = self.execute(module, np)
            self.assertEqual(status, 0)
            self.assertEqual(stdout.count(message), np)
            self.assertEqual(stderr, "")


class TestRunCommand(BaseTestRun):
    def execute(self, command, np=1):
        return execute(np, "-c", shlex.quote(command))

    def testArgv0(self):
        command = "import sys; print(sys.argv[0], flush=True)"
        status, stdout, stderr = self.execute(command, 1)
        self.assertEqual(status, 0)
        self.assertEqual(stdout.strip(), "-c")
        self.assertEqual(stderr.strip(), "")

    def testSuccess(self):
        command = "from mpi4py import MPI"
        for np in (1, 2):
            status, stdout, stderr = self.execute(command, np)
            self.assertEqual(status, 0)
            self.assertEqual(stdout, "")
            self.assertEqual(stderr, "")

    def testException(self):
        command = "; ".join((
            "from mpi4py import MPI",
            "comm = MPI.COMM_WORLD",
            "comm.Barrier()",
            "comm.Barrier()",
            "comm.Get_rank() == {} and (1/0)",
            "comm.Barrier()",
        ))
        excmess = "ZeroDivisionError:"
        for np in (1, 2):
            for rank in range(np):
                status, stdout, stderr = self.execute(command.format(rank), np)
                self.assertFailure(status, 1)
                self.assertMPIAbort(stdout, stderr, excmess)


if __name__ == "__main__":
    unittest.main()



import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


def create_topo_comms(comm):
    size = comm.Get_size()
    rank = comm.Get_rank()
    # Cartesian
    n = int(size**1 / 2.0)
    m = int(size**1 / 3.0)
    if m * m * m == size:
        dims = [m, m, m]
    elif n * n == size:
        dims = [n, n]
    else:
        dims = [size]
    periods = [True] * len(dims)
    yield comm.Create_cart(dims, periods=periods)
    # Graph
    index, edges = [0], []
    for i in range(size):
        pos = index[-1]
        index.append(pos + 2)
        edges.extend(((i - 1) % size, (i + 1) % size))
    yield comm.Create_graph(index, edges)
    # Dist Graph
    sources = [(rank - 2) % size, (rank - 1) % size]
    destinations = [(rank + 1) % size, (rank + 2) % size]
    yield comm.Create_dist_graph_adjacent(sources, destinations)


def get_neighbors_count(comm):
    topo = comm.Get_topology()
    if topo == MPI.CART:
        ndim = comm.Get_dim()
        return 2 * ndim, 2 * ndim
    if topo == MPI.GRAPH:
        rank = comm.Get_rank()
        nneighbors = comm.Get_neighbors_count(rank)
        return nneighbors, nneighbors
    if topo == MPI.DIST_GRAPH:
        indeg, outdeg, _w = comm.Get_dist_neighbors_count()
        return indeg, outdeg
    return 0, 0


def StartWaitFree(request):
    request.Start()
    request.Wait()
    request.Free()


class BaseTestCCONghBuf:
    #
    COMM = MPI.COMM_NULL

    def testNeighborAllgather(self):
        for comm in create_topo_comms(self.COMM):
            rsize, _ssize = get_neighbors_count(comm)
            for array, typecode in arrayimpl.loop():
                for v in range(3):
                    check = arrayimpl.scalar(v)
                    sbuf = array(v, typecode, 3)
                    rbuf = array(-1, typecode, (rsize, 3))
                    StartWaitFree(
                        comm.Neighbor_allgather_init(
                            sbuf.as_mpi(), rbuf.as_mpi()
                        )
                    )
                    for value in rbuf.flat:
                        self.assertEqual(value, check)
                    sbuf = array(v, typecode, 3)
                    rbuf = array(-1, typecode, (rsize, 3))
                    StartWaitFree(
                        comm.Neighbor_allgatherv_init(
                            sbuf.as_mpi_c(3), rbuf.as_mpi_c(3)
                        )
                    )
                    for value in rbuf.flat:
                        self.assertEqual(value, check)
            comm.Free()

    def testNeighborAlltoall(self):
        for comm in create_topo_comms(self.COMM):
            rsize, ssize = get_neighbors_count(comm)
            for array, typecode in arrayimpl.loop():
                for v in range(3):
                    check = arrayimpl.scalar(v)
                    sbuf = array(v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    StartWaitFree(
                        comm.Neighbor_alltoall_init(
                            sbuf.as_mpi(), rbuf.as_mpi_c(3)
                        )
                    )
                    for value in rbuf.flat:
                        self.assertEqual(value, check)
                    sbuf = array(v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    StartWaitFree(
                        comm.Neighbor_alltoall_init(
                            sbuf.as_mpi(), rbuf.as_mpi()
                        )
                    )
                    for value in rbuf.flat:
                        self.assertEqual(value, check)
                    sbuf = array(v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    StartWaitFree(
                        comm.Neighbor_alltoallv_init(
                            sbuf.as_mpi_c(3), rbuf.as_mpi_c(3)
                        )
                    )
                    for value in rbuf.flat:
                        self.assertEqual(value, check)
            comm.Free()

    def testNeighborAlltoallw(self):
        for comm in create_topo_comms(self.COMM):
            rsize, ssize = get_neighbors_count(comm)
            for array, typecode in arrayimpl.loop():
                for n in range(1, 4):
                    for v in range(3):
                        check = arrayimpl.scalar(v)
                        sbuf = array(v, typecode, (ssize, n))
                        rbuf = array(-1, typecode, (rsize, n))
                        sdt, rdt = sbuf.mpidtype, rbuf.mpidtype
                        sdsp = list(
                            range(0, ssize * n * sdt.extent, n * sdt.extent)
                        )
                        rdsp = list(
                            range(0, rsize * n * rdt.extent, n * rdt.extent)
                        )
                        smsg = [
                            sbuf.as_raw(),
                            ([n] * ssize, sdsp),
                            [sdt] * ssize,
                        ]
                        rmsg = (
                            rbuf.as_raw(),
                            ([n] * rsize, rdsp),
                            [rdt] * rsize,
                        )
                        StartWaitFree(comm.Neighbor_alltoallw_init(smsg, rmsg))
                        for value in rbuf.flat:
                            self.assertEqual(value, check)
            comm.Free()


class TestCCONghBufSelf(BaseTestCCONghBuf, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestCCONghBufWorld(BaseTestCCONghBuf, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestCCONghBufSelfDup(TestCCONghBufSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


class TestCCONghBufWorldDup(TestCCONghBufWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


try:
    StartWaitFree(MPI.COMM_SELF.Barrier_init())
except NotImplementedError:
    unittest.disable(BaseTestCCONghBuf, "mpi-coll-persist")


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI

try:
    import cffi
except ImportError:
    cffi = None


@unittest.skipIf(cffi is None, "cffi")
class TestCFFI(unittest.TestCase):
    #
    mpitypes = [
        MPI.Datatype,
        MPI.Request,
        MPI.Info,
        MPI.Errhandler,
        MPI.Session,
        MPI.Group,
        MPI.Win,
        MPI.Op,
        MPI.File,
        MPI.Message,
        MPI.Comm,
    ]

    objects = [
        MPI.DATATYPE_NULL,
        MPI.INT,
        MPI.DOUBLE,
        MPI.REQUEST_NULL,
        MPI.INFO_NULL,
        MPI.INFO_ENV,
        MPI.ERRHANDLER_NULL,
        MPI.ERRORS_RETURN,
        MPI.ERRORS_ARE_FATAL,
        MPI.GROUP_NULL,
        MPI.GROUP_EMPTY,
        MPI.WIN_NULL,
        MPI.OP_NULL,
        MPI.SUM,
        MPI.MIN,
        MPI.MAX,
        MPI.FILE_NULL,
        MPI.MESSAGE_NULL,
        MPI.MESSAGE_NO_PROC,
        MPI.COMM_NULL,
        MPI.COMM_SELF,
        MPI.COMM_WORLD,
    ]

    def testHandleAddress(self):
        ffi = cffi.FFI()
        typemap = {
            ffi.sizeof("int"): "int",
            ffi.sizeof("void*"): "void*",
        }

        def typename(t):
            return t.__name__.rsplit(".", 1)[-1]

        for tp in self.mpitypes:
            handle_t = typemap[MPI._sizeof(tp)]
            mpi_t = "MPI_" + typename(tp)
            ffi.cdef(f"typedef {handle_t} {mpi_t};")
        for obj in self.objects:
            if isinstance(obj, MPI.Comm):
                mpi_t = "MPI_Comm"
            else:
                mpi_t = "MPI_" + typename(type(obj))
            oldobj = obj
            newobj = type(obj)()
            handle_old = ffi.cast(mpi_t + "*", MPI._addressof(oldobj))
            handle_new = ffi.cast(mpi_t + "*", MPI._addressof(newobj))
            handle_new[0] = handle_old[0]
            self.assertEqual(oldobj, newobj)

    def testHandleValue(self):
        ffi = cffi.FFI()
        typemap = {
            ffi.sizeof("uint32_t"): "uint32_t",
            ffi.sizeof("uint64_t"): "uint64_t",
        }
        for obj in self.objects:
            uintptr_t = typemap[MPI._sizeof(obj)]
            handle = ffi.cast(uintptr_t + "*", MPI._addressof(obj))[0]
            self.assertEqual(handle, MPI._handleof(obj))


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI


class TestErrorCode(unittest.TestCase):
    #
    errorclasses = [
        item[1] for item in vars(MPI).items() if item[0].startswith("ERR_")
    ]
    errorclasses.insert(0, MPI.SUCCESS)
    while MPI.ERR_LASTCODE in errorclasses:
        errorclasses.remove(MPI.ERR_LASTCODE)

    def testGetErrorClass(self):
        is_openmpi = unittest.is_mpi("openmpi")
        self.assertEqual(self.errorclasses[0], 0)
        for ierr in self.errorclasses:
            errcls = MPI.Get_error_class(ierr)
            self.assertGreaterEqual(errcls, MPI.SUCCESS)
            self.assertLessEqual(errcls, MPI.ERR_LASTCODE)
            if is_openmpi and errcls == MPI.ERR_UNKNOWN and errcls != ierr:
                continue
            self.assertEqual(errcls, ierr)

    def testGetErrorStrings(self):
        for ierr in self.errorclasses:
            errstr = MPI.Get_error_string(ierr)
            self.assertGreater(len(errstr), 0)

    def testException(self):
        success = MPI.Exception(MPI.SUCCESS)
        lasterr = MPI.Exception(MPI.ERR_LASTCODE)
        for ierr in self.errorclasses:
            errstr = MPI.Get_error_string(ierr)
            errcls = MPI.Get_error_class(ierr)
            errexc = MPI.Exception(ierr)
            self.assertEqual(errexc.Get_error_code(), ierr)
            self.assertEqual(errexc.Get_error_class(), errcls)
            self.assertEqual(errexc.Get_error_string(), errstr)
            self.assertEqual(errexc.error_code, ierr)
            self.assertEqual(errexc.error_class, errcls)
            self.assertEqual(errexc.error_string, errstr)
            self.assertEqual(repr(errexc), f"MPI.Exception({ierr})")
            self.assertEqual(str(errexc), errstr)
            self.assertEqual(int(errexc), ierr)
            self.assertEqual(hash(errexc), hash(errexc.error_code))
            self.assertTrue(bool(errexc == ierr))
            self.assertTrue(bool(errexc == errexc))
            self.assertFalse(bool(errexc != ierr))
            self.assertFalse(bool(errexc != errexc))
            self.assertTrue(bool(success <= ierr <= lasterr))
            self.assertTrue(bool(success <= errexc <= lasterr))
            self.assertTrue(bool(errexc >= ierr))
            self.assertTrue(bool(errexc >= success))
            self.assertTrue(bool(lasterr >= ierr))
            self.assertTrue(bool(lasterr >= errexc))
            if errexc == success:
                self.assertFalse(errexc)
            else:
                self.assertTrue(errexc)
                self.assertTrue(bool(errexc > success))
                self.assertTrue(bool(success < errexc))
        exc = MPI.Exception(MPI.SUCCESS - 1)
        self.assertEqual(exc, MPI.ERR_UNKNOWN)
        exc = MPI.Exception(MPI.ERR_LASTCODE + 1)
        self.assertEqual(exc, MPI.ERR_LASTCODE + 1)

    @unittest.skipMPI("openmpi(<1.10.0)")
    def testAddErrorClass(self):
        try:
            errclass = MPI.Add_error_class()
        except NotImplementedError:
            self.skipTest("mpi-add_error_class")
        self.assertGreaterEqual(errclass, MPI.ERR_LASTCODE)
        try:
            MPI.Remove_error_class(errclass)
        except NotImplementedError:
            pass

    @unittest.skipMPI("openmpi(<1.10.0)")
    def testAddErrorCode(self):
        try:
            errcode = MPI.Add_error_code(MPI.ERR_OTHER)
        except NotImplementedError:
            self.skipTest("mpi-add_error_code")
        try:
            MPI.Remove_error_code(errcode)
        except NotImplementedError:
            pass

    @unittest.skipMPI("openmpi(<1.10.0)")
    def testAddErrorClassCodeString(self):
        MPI.COMM_WORLD.Get_attr(MPI.LASTUSEDCODE)
        try:
            errclass = MPI.Add_error_class()
        except NotImplementedError:
            self.skipTest("mpi-add_error_class")
        lastused = MPI.COMM_WORLD.Get_attr(MPI.LASTUSEDCODE)
        self.assertEqual(errclass, lastused)
        errstr = MPI.Get_error_string(errclass)
        self.assertEqual(errstr, "")
        MPI.Add_error_string(errclass, "error class")
        self.assertEqual(MPI.Get_error_string(errclass), "error class")
        errcode1 = MPI.Add_error_code(errclass)
        errstr = MPI.Get_error_string(errcode1)
        self.assertEqual(errstr, "")
        MPI.Add_error_string(errcode1, "error code 1")
        self.assertEqual(MPI.Get_error_class(errcode1), errclass)
        self.assertEqual(MPI.Get_error_string(errcode1), "error code 1")
        errcode2 = MPI.Add_error_code(errclass)
        errstr = MPI.Get_error_string(errcode2)
        self.assertEqual(errstr, "")
        MPI.Add_error_string(errcode2, "error code 2")
        self.assertEqual(MPI.Get_error_class(errcode2), errclass)
        self.assertEqual(MPI.Get_error_string(errcode2), "error code 2")
        #
        with self.catchNotImplementedError(4, 1):
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_class(errclass)
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_code(errcode2)
            MPI.Remove_error_string(errcode2)
            self.assertEqual(MPI.Get_error_string(errcode2), "")
            MPI.Remove_error_code(errcode2)
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_code(errcode2)
            #
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_class(errclass)
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_code(errcode1)
            MPI.Remove_error_string(errcode1)
            self.assertEqual(MPI.Get_error_string(errcode1), "")
            MPI.Remove_error_code(errcode1)
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_code(errcode1)
            #
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_class(errclass)
            MPI.Remove_error_string(errclass)
            self.assertEqual(MPI.Get_error_string(errclass), "")
            MPI.Remove_error_class(errclass)
            with self.assertRaises(MPI.Exception):
                MPI.Remove_error_class(errclass)
        #
        try:
            MPI.Remove_error_class(0)
            self.fail("expected Exception")
        except (MPI.Exception, NotImplementedError):
            pass
        try:
            MPI.Remove_error_code(0)
            self.fail("expected Exception")
        except (MPI.Exception, NotImplementedError):
            pass
        try:
            MPI.Remove_error_string(0)
            self.fail("expected Exception")
        except (MPI.Exception, NotImplementedError):
            pass


if __name__ == "__main__":
    unittest.main()
